@@ -2,7 +2,7 @@
 /*
  * FCRON - periodic command scheduler 
  *
- *  Copyright 2000-2001 Thibault Godouet <fcron@free.fr>
+ *  Copyright 2000-2004 Thibault Godouet <fcron@free.fr>
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -22,12 +22,18 @@
  *  `LICENSE' that comes with the fcron source distribution.
  */
 
- /* $Id: conf.c,v 1.40 2001/02/27 20:50:03 thib Exp $ */
+ /* $Id: conf.c,v 1.66 2003/12/25 22:43:57 thib Exp thib $ */
 
 #include "fcron.h"
 
-int read_file(const char *file_name, CF *cf);
-char *read_str(FILE *f, char *buf, int max);
+#include "conf.h"
+#include "database.h"
+
+int read_file(const char *file_name, cf_t *cf);
+int add_line_to_file(cl_t *cl, cf_t *cf, uid_t runas, char *runas_str,
+		     time_t t_save);
+int read_strn(int fd, char **str, short int size);
+int read_type(int fd, short int *type, short int *size);
 void synchronize_file(char *file_name);
 
 
@@ -43,7 +49,7 @@ reload_all(const char *dir_name)
     /* save all current configuration, remove it from the memory,
      * and reload from dir_name */
 {
-    CF *f = NULL;
+    cf_t *f = NULL;
 
     explain("Removing current configuration from memory");
 
@@ -80,7 +86,7 @@ synchronize_dir(const char *dir_name)
     struct dirent *den;
 
     if ( strcmp(dir_name, ".") == 0 )
-	explain("updating configuration from " FCRONTABS );
+	explain("updating configuration from %s", fcrontabs);
     else
 	explain("updating configuration from %s", dir_name);
 
@@ -124,17 +130,7 @@ synchronize_dir(const char *dir_name)
 
     /* proceed to adds or removes */
 
-    /* begin by adding normal files, if any, to database */
-    for (list_cur = file_list; list_cur; list_cur = list_cur->next ) {
-	if ( getpwnam(list_cur->str) ) {
-	    explain("adding file %s", list_cur->str);	    
-	    synchronize_file(list_cur->str);
-	}
-	else
-	    error("ignoring file \"%s\" : not in passwd file.", list_cur->str);
-    }
-
-    /* then remove files which are no longer wanted */
+    /* begin by removing files which are no longer wanted */
     for (list_cur = rm_list; list_cur; list_cur = list_cur->next ) {
 	explain("removing file %s", list_cur->str + 3);
 	delete_file(list_cur->str + 3);  /* len("rm.") = 3 */
@@ -144,9 +140,28 @@ synchronize_dir(const char *dir_name)
 	    error_e("Could not remove %s", list_cur->str);
     }
     
+    /* then add normal files, if any, to database */
+    for (list_cur = file_list; list_cur; list_cur = list_cur->next ) {
+	if ( getpwnam(list_cur->str) 
+#ifdef SYSFCRONTAB
+	     || strcmp(list_cur->str, SYSFCRONTAB) == 0
+#endif
+	    ) {
+	    explain("adding file %s", list_cur->str);	    
+	    synchronize_file(list_cur->str);
+	}
+	else
+	    error("ignoring file \"%s\" : not in passwd file.", list_cur->str);
+    }
+
     /* finally add new files */
     for (list_cur = new_list; list_cur; list_cur = list_cur->next ) {
-	if ( getpwnam(list_cur->str + 4) ) { /* len("new.") = 4 */
+	 /* len("new.") = 4 : */
+	if ( getpwnam(list_cur->str + 4)
+#ifdef SYSFCRONTAB
+	     || strcmp(list_cur->str + 4, SYSFCRONTAB) == 0
+#endif
+	    ) {
 	    explain("adding new file %s", list_cur->str + 4);
 	    synchronize_file(list_cur->str);  
 	}
@@ -190,7 +205,7 @@ void
 synchronize_file(char *file_name)
 {
 
-    CF *cur_f = NULL;
+    cf_t *cur_f = NULL;
     char *user = NULL;
 
     if (strchr(file_name, '.') != NULL ) {
@@ -198,7 +213,7 @@ synchronize_file(char *file_name)
 	 * version in database in order to keep a maximum of fields
 	 * (cl_nextexe) to their current value */
 
-	CF *prev = NULL;
+	cf_t *prev = NULL;
 
 	/* set user name  */
 	/* we add 4 to file_name pointer because of the "new."
@@ -214,9 +229,9 @@ synchronize_file(char *file_name)
 	if (cur_f != NULL) {
 	    /* an old version of this file exist in database */
 
-	    CF *old = NULL;
-	    CL *old_l = NULL;
-	    CL *new_l = NULL;
+	    cf_t *old = NULL;
+	    cl_t *old_l = NULL;
+	    cl_t *new_l = NULL;
 	    /* size used when comparing two line : 
 	     * it's the size of all time table (mins, days ...) */
 	    const size_t size=( bitstr_size(60) + bitstr_size(24) + 
@@ -226,10 +241,9 @@ synchronize_file(char *file_name)
 	    old = cur_f;
 
 	    /* load new file */
-	    Alloc(cur_f, CF);
+	    Alloc(cur_f, cf_t);
 	    if ( read_file(file_name, cur_f) != 0 ) {
 		/* an error as occured */
-		free(cur_f);
 		return;
 	    }
 
@@ -264,7 +278,16 @@ synchronize_file(char *file_name)
 			
 			if ( new_l->cl_runfreq == old_l->cl_runfreq ) 
 			    new_l->cl_remain = old_l->cl_remain;
-			new_l->cl_nextexe = old_l->cl_nextexe;
+			/* check if there is a change about the tz diff */
+			if ( (new_l->cl_file->cf_tzdiff != 
+			      old_l->cl_file->cf_tzdiff) &&
+			     (old_l->cl_nextexe - old_l->cl_file->cf_tzdiff
+			      + new_l->cl_file->cf_tzdiff > now) )
+			    new_l->cl_nextexe = old_l->cl_nextexe
+				- old_l->cl_file->cf_tzdiff
+				+ new_l->cl_file->cf_tzdiff;
+			else
+			    new_l->cl_nextexe = old_l->cl_nextexe;
 			insert_nextexe(new_l);
 
 			if (debug_opt) {
@@ -302,11 +325,10 @@ synchronize_file(char *file_name)
 	    /* no old version exist in database : load this file
 	     * as a normal file, but change its name */
 	
-	    Alloc(cur_f, CF);
+	    Alloc(cur_f, cf_t);
 
 	    if ( read_file(file_name, cur_f) != 0 ) {
 		/* an error as occured */
-		free(cur_f);
 		return;
 	    }
 
@@ -327,11 +349,10 @@ synchronize_file(char *file_name)
     else {
 	/* this is a normal file */
 	
-	Alloc(cur_f, CF);
+	Alloc(cur_f, cf_t);
 
 	if ( read_file(file_name, cur_f) != 0 ) {
 	    /* an error as occured */
-	    free(cur_f);
 	    return;
 	}
 
@@ -344,279 +365,523 @@ synchronize_file(char *file_name)
 }
 
 
-char *
-read_str(FILE *f, char *buf, int max)
-    /* return a pointer to string read from file f
-     * if it is non-zero length */
+int
+read_strn(int fd, char **str, short int size)
+/* read a "size"-length string in a binary fcrontab file */
 {
-    int i;
+    if ( (*str = calloc(size + 1, sizeof(char))) == NULL )
+	goto err;
+    
+    if ( read(fd, *str, size) < size )
+	goto err;
+    (*str)[size] = '\0';
+    return OK;
 
-    for (i = 0; i < max; i++)
-	if ( (buf[i] = fgetc(f)) == '\0')
-	    break;
-    buf[max-1] = '\0';
+  err:
+    if (*str)
+	free(*str);
+    *str = NULL;
+    return ERR;
+    
+}
 
-    if ( strlen(buf) == 0 )
-	return NULL;
-    else
-	return strdup2(buf);
+int
+read_type(int fd, short int *type, short int *size)
+/* read the type and size of the next field in a binary fcrontab file */
+{
+    if ( read(fd, type, sizeof(short int)) < sizeof(short int) )
+	goto err;
+    if ( read(fd, size, sizeof(short int)) < sizeof(short int) )
+	goto err;
+    
+    return OK;
+
+  err:
+    return ERR;
 
 }
 
 
+/* macros for read_file() */
+/* read "size" bytes from file "ff", put them in "to", and check for errors */
+#define Read(TO, SIZE, ERR_STR) \
+        { \
+          if ( read(fileno(ff), &(TO), SIZE) < SIZE ) { \
+            error_e(ERR_STR); \
+	    goto err; \
+	  } \
+        }
+
+#define Read_strn(TO, SIZE, ERR_STR) \
+        { \
+          if ( read_strn(fileno(ff), &(TO), SIZE) != OK ) { \
+            error_e(ERR_STR); \
+	    goto err; \
+	  } \
+        }
+
 int
-read_file(const char *file_name, CF *cf)
+read_file(const char *file_name, cf_t *cf)
     /* read a formated fcrontab.
-       return 1 on error, 0 otherwise */
+       return ERR on error, OK otherwise */
 {
     FILE *ff = NULL;
-    CL *cl = NULL;
+    cl_t *cl = NULL;
     env_t *env = NULL;
     char buf[LINE_LEN];
+    long int bufi = 0;
     time_t t_save = 0;
-    time_t slept = 0;
     uid_t runas = 0;
     char *runas_str = NULL;
     struct stat file_stat;
     struct passwd *pass = NULL;
+    short int type = 0, size = 0;
+    int rc;
+#ifdef CONFIG_FLASK
+    int flask_enabled = is_flask_enabled();
+    struct security_query qry;
+    struct security_response rsp;
+    int retval;
+    const char *user_name;
+#endif
 
     /* open file */
     if ( (ff = fopen(file_name, "r")) == NULL ) {
-	error_e("Could not read %s", file_name);
-	return 1;
+	warn_e("Could not read %s (may have just been removed)", file_name);
+	goto err;
     }
 
     /* check if this file is owned by root : otherwise, all runas fields
      * of this field should be set to the owner */
-    if ( fstat(fileno(ff), &file_stat) != 0 ) {
+#ifdef CONFIG_FLASK
+    if(flask_enabled)
+	rc = fstat_secure(fileno(ff), &file_stat, &cf->cf_file_sid);
+    else
+#endif
+    rc = fstat(fileno(ff), &file_stat);
+    if ( rc != 0 ) {
 	error_e("Could not stat %s", file_name);
-	return 1;
+	goto err;
     }
 
     if ( strncmp(file_name,"new.", 4) == 0 ) {
-	if ( file_stat.st_uid == 0 )
-	    /* file is owned by root : no test needed : set runas to 0 */
-	    runas = 0;
+	if ( file_stat.st_uid == ROOTUID ) {
+	    /* file is owned by root : no test needed : set runas to ROOTUID */
+	    runas = ROOTUID;
+	}
 	else {
 	    /* this is a standard user's new fcrontab : set the runas field to
 	     * the owner of the file */
 	    runas = file_stat.st_uid;
 	    if ( (pass = getpwuid(file_stat.st_uid)) == NULL ) {
 		error_e("Could not getpwuid(%d)", file_stat.st_uid);
-		return 1;
+		goto err;
 	    }
-	    /* set cf_user field */
-	    runas_str = cf->cf_user = strdup2(pass->pw_name);
+	    runas_str = strdup2(pass->pw_name);
 	}
+	cf->cf_user = strdup2(file_name + 4);
     }
     else {
-	if ( file_stat.st_uid == 0 )
+	if(!cf->cf_user)
+	    cf->cf_user = strdup2(file_name);
+	if ( file_stat.st_uid == ROOTUID ) {
 	    /* file is owned by root : either this file has already been parsed
-	     * at least once by fcron, either it is the root's fcrontab */
-	    runas = 0;
+	     * at least once by fcron, or it is root's fcrontab */
+	    runas = ROOTUID;
+	}
 	else {
 	    error("Non-new file %s owned by someone else than root",file_name);
-	    return 1;
+	    goto err;
 	}
     }
 
+#ifdef CONFIG_FLASK
+    /*
+     * Since crontab files are not directly executed,
+     * crond must ensure that the crontab file has
+     * a context that is appropriate for the context of
+     * the user cron job.  It performs an entrypoint
+     * permission check for this purpose.
+     */
+    if(!strcmp(cf->cf_user, SYSFCRONTAB))
+	user_name = "system_u";
+    else
+	user_name = cf->cf_user;
+    if(get_default_sid(user_name, 0, &cf->cf_user_sid))
+	error_e("NO SID for user \"%s\"", cf->cf_user_sid);
+    qry.ssid = cf->cf_user_sid;
+    qry.tsid = cf->cf_file_sid;
+    qry.tclass = SECCLASS_FILE;
+    qry.requested = FILE__ENTRYPOINT;
+    retval = security_compute_av(&qry, &rsp);
+    if(retval || ((qry.requested & rsp.allowed) != qry.requested)) {
+	syslog(LOG_ERR, "ENTRYPOINT FAILED for \"%s\" (SID %u) for file SID %u"
+	       , cf->cf_user, cf->cf_user_sid, cf->cf_file_sid);
+	goto err;
+    }
+#endif
+
     debug("User %s Entry", file_name);
     bzero(buf, sizeof(buf));
 
-    /* get version of fcrontab file: it permit to daemon not to load
+    /* get version of fcrontab file: it permits to daemon not to load
      * a file which he won't understand the syntax, for exemple
      * a file using a depreciated format generated by an old fcrontab,
      * if the syntax has changed */
-    if (fgets(buf, sizeof(buf), ff) == NULL ||
-	strncmp(buf, "fcrontab-"FILEVERSION"\n",
-		sizeof("fcrontab-"FILEVERSION"\n")) != 0) {
-
+    if ( read_type(fileno(ff), &type, &size) != OK || type != S_HEADER_T ||
+	 read(fileno(ff), &bufi, size) < size || bufi != S_FILEVERSION ) {
 	error("File %s is not valid: ignored.", file_name);
-	error("Maybe this file has been generated by an old version"
-	      " of fcron. In this case, you should install it again"
-	      " using fcrontab.");
-	return 1;
+	error("This file may have been generated by an old version of fcron.");
+	error("In that case, you should try to use the converter given in the "
+	      "source package, or install it again using fcrontab.");
+	goto err;
     }    
 
-    if ( runas == 0 ) {
-	/* get the owner's name */
-	if ((cf->cf_user = read_str(ff, buf, sizeof(buf))) == NULL) {
-	    error("Cannot read user's name : file ignored");
-	    return 1;
-	}
+    if ( read_type(fileno(ff), &type, &size) != OK || type != S_USER_T ) {
+	error("Invalid binary fcrontab (no USER field)");
+	goto err;
+    }
+    /* get the owner's name */
+    /* we set cf->cf_user before for SE Linux, so we need to free it here */
+    free(cf->cf_user);
+    if ( read_strn(fileno(ff), &cf->cf_user, size) != OK ) {
+	error("Cannot read user's name : file ignored");
+	goto err;
+    }
+    if ( runas != ROOTUID ) {
+	/* we use file owner's name for more security (see above) */
+	/* free the value obtained by read_strn() (we need to read it anyway
+	 * to set the file ptr to the next thing to read) */
+	free(cf->cf_user);
+	cf->cf_user = runas_str;
     } 
-    else 
-	/* read the user name but ignore it */
-	read_str(ff, buf, sizeof(buf));
 
     /* get the time & date of the saving */
     /* a new file generated by fcrontab has t_save set to 0 */
-    if ( fscanf(ff, "%ld", &t_save) != 1 )
+    if ( read_type(fileno(ff), &type, &size) != OK || type != S_TIMEDATE_T 
+	 || read(fileno(ff), &t_save, size) < size ) {
 	error("could not get time and date of saving");
-	
-    slept = now - t_save;
-
-
-    /* read env variables */
-    Alloc(env, env_t);
-    while( (env->e_val = read_str(ff, buf, sizeof(buf))) != NULL ) {
-	debug("  Env: \"%s\"", env->e_val );
-	
-	env->e_next = cf->cf_env_base;
-	cf->cf_env_base = env;
-	Alloc(env, env_t);
+	goto err;
     }
-    /* free last alloc : unused */
-    free(env);
 	
-    /* read lines */
-    Alloc(cl, CL);
-    while ( fread(cl, sizeof(CL), 1, ff) == 1 ) {
+    Alloc(cl, cl_t);
+    /* main loop : read env variables, and lines */
+    while ( read_type(fileno(ff), &type, &size) == OK ) {
+	/* action is determined by the type of the field */
+	switch ( type ) {
 
-	if ((cl->cl_shell = read_str(ff, buf, sizeof(buf))) == NULL) {
-	    error("Line is not valid (empty shell command) : ignored");
-	    continue;
-	}
-	if ((cl->cl_runas = read_str(ff, buf, sizeof(buf))) == NULL) {
-	    error("Line is not valid (empty runas field) : ignored");
-	    continue;
-	}
-	if ((cl->cl_mailto = read_str(ff, buf, sizeof(buf))) == NULL) {
-	    error("Line is not valid (empty mailto field) : ignored");
-	    continue;
-	}
+	case S_ENVVAR_T:
+	    /* read a env variable and add it to the env var list */
+	    Alloc(env, env_t);
+	    Read_strn(env->e_val, size, "Error while reading env var");
+	    /* Read_strn go to "err" on error */
+	    debug("  Env: \"%s\"", env->e_val );
+	    env->e_next = cf->cf_env_base;
+	    cf->cf_env_base = env;	    
+	    break;
 
-	/* set runas field if necessary (to improve security) */
-	if (runas > 0) {
-	    if (strcmp(cl->cl_runas, runas_str) != 0)
-		warn("warning: runas(%s) is not owner (%s): overridden.",
-		     cl->cl_runas, runas_str);
-	    free(cl->cl_runas);
-	    cl->cl_runas = strdup2(runas_str);
-	}
+	case S_TZDIFF_T:
+	    /* time diff between local (real) and system hour */
+	    Read(bufi, size, "Error while reading tzdiff field");
+	    cf->cf_tzdiff = (signed char) bufi;
+	    break;
 
-	/* we need that here because the user's name (contained in the
-	 * struct CF may be required */
-	cl->cl_file = cf;
+	case S_SHELL_T:
+	    Read_strn(cl->cl_shell, size, "Error while reading shell field");
+	    break;
 
-	/* check if the task has not been stopped during execution and insert
-	 * in lavg or serial queues the jobs which was in one at fcron's stops
-	 */
-	if (cl->cl_numexe > 0) {
-	    cl->cl_numexe = 0;
-	    if ( ! is_strict(cl->cl_option) ) {
-		if ( is_lavg(cl->cl_option) )
-		    add_lavg_job(cl);
-		else if ( is_serial(cl->cl_option) 
-			  || is_serial_once(cl->cl_option) )
-		    add_serial_job(cl);
-		else {
-		    /* job has been stopped during execution :
-		     * launch it again */
-		    warn("job %s did not finish : running it again.",
-			 cl->cl_shell);
-		    set_serial_once(cl->cl_option);
-		    add_serial_job(cl);
-		}
+	case S_RUNAS_T:
+	    Read_strn(cl->cl_runas, size, "Error while reading runas field");
+	    break;
+
+	case S_MAILTO_T:
+	    Read_strn(cl->cl_mailto, size, "Error while reading mailto field");
+	    break;
+
+	case S_NEXTEXE_T:
+	    Read(bufi, size, "Error while reading nextexe field");
+	    cl->cl_nextexe = (time_t) bufi;
+	    break;
+
+	case S_FIRST_T:
+	    Read(bufi, size, "Error while reading first field");
+	    cl->cl_first = (time_t) bufi;
+	    break;
+
+	case S_OPTION_T:
+	    if (size < OPTION_SIZE)
+		/* set the options not defined in the savefile to default */
+		set_default_opt(cl->cl_option);
+	    Read(cl->cl_option, size, "Error while reading option field");
+	    break;
+
+	case S_NUMEXE_T:
+	    Read(cl->cl_numexe, size, "Error while reading numexe field");
+	    break;
+
+	case S_LAVG_T:
+	    Read(cl->cl_lavg, size, "Error while reading lavg field");
+	    break;
+
+	case S_UNTIL_T:
+	    Read(bufi, size, "Error while reading until field");
+	    cl->cl_until = (time_t) bufi;
+	    break;
+
+	case S_NICE_T:
+	    Read(cl->cl_nice, size, "Error while reading nice field");
+	    break;
+
+	case S_RUNFREQ_T:
+	    Read(bufi, size, "Error while reading runfreq field");
+	    cl->cl_runfreq = (unsigned short) bufi;
+	    break;
+
+	case S_REMAIN_T:
+	    Read(bufi, size, "Error while reading remain field");
+	    cl->cl_remain = (unsigned short) bufi;
+	    break;
+
+	case S_TIMEFREQ_T:
+	    Read(bufi, size, "Error while reading timefreq field");
+	    cl->cl_timefreq = (time_t) bufi;
+	    break;
+
+	case S_MINS_T:
+	    Read(cl->cl_mins, size, "Error while reading mins field");
+	    break;
+
+	case S_HRS_T:
+	    Read(cl->cl_hrs, size, "Error while reading hrs field");
+	    break;
+
+	case S_DAYS_T:
+	    Read(cl->cl_days, size, "Error while reading days field");
+	    break;
+
+	case S_MONS_T:
+	    Read(cl->cl_mons, size, "Error while reading mons field");
+	    break;
+
+	case S_DOW_T:
+	    Read(cl->cl_dow, size, "Error while reading dow field");
+	    break;
+
+	case S_ENDLINE_T:
+	    if (add_line_to_file(cl, cf, runas, runas_str, t_save) == 0)
+		Alloc(cl, cl_t);
+	    break;
+
+	    /* default case in "switch(type)" */
+	default:
+	    error("Error while loading %s : unknown field type %d (ignored)",
+		  file_name, type);
+	    /* skip the data corresponding to the unknown field */
+	    {
+		/* we avoid using fseek(), as it seems not to work correctly
+		 * on some systems when we use read() on the FILE stream */
+		int i;
+		for (i = 0; i < size; i++)
+		    if ( getc(ff) == EOF )
+			goto err;
 	    }
 	}
-
-	if ( is_td(cl->cl_option) ) {
-    
-	    /* set the time and date of the next execution  */
-	    if ( cl->cl_nextexe <= now ) {
-		if ( cl->cl_nextexe == 0 )
-		    /* the is a line from a new file */
-		    set_next_exe(cl, NO_GOTO);		    
-		else if(cl->cl_runfreq == 1 && is_notice_notrun(cl->cl_option))
-		    set_next_exe_notrun(cl, SYSDOWN);
-		else if ( is_bootrun(cl->cl_option) && t_save != 0 
-			  && cl->cl_runfreq != 1) {
-		    if ( cl->cl_remain > 0 && --cl->cl_remain > 0 ) {
-			debug("    cl_remain: %d", cl->cl_remain);
-		    }
-		    else {
-			/* run bootrun jobs */
-			cl->cl_remain = cl->cl_runfreq;
-			debug("   boot-run %s", cl->cl_shell);
-			if ( ! is_lavg(cl->cl_option) ) {
-			    set_serial_once(cl->cl_option);
-			    add_serial_job(cl);
-			}
-			else
-			    add_lavg_job(cl);			    
-		    }
-		    set_next_exe(cl, STD);
-		}
-		else {
-		    if ( is_notice_notrun(cl->cl_option) ) {
-			/* set next exe and mail user */
-			struct tm *since2 = localtime(&cl->cl_nextexe), since;
-			memcpy(&since, since2, sizeof(since));
-			set_next_exe(cl, NO_GOTO);
-			mail_notrun(cl, SYSDOWN, &since);
-		    } 
-		    else
-			set_next_exe(cl, NO_GOTO);
-		}
-	    }
-	    else
-		/* value of nextexe is valid : just insert line in queue */
-		insert_nextexe(cl);
-	} else {  /* is_td(cl->cl_option) */
-	    /* standard @-lines */
-	    if ( cl->cl_timefreq < 60 ) {
-		error("Invalid timefreq for %s: set to 1 day", cl->cl_shell);
-		cl->cl_timefreq = 3600*24;
-	    }
-	    cl->cl_nextexe += slept;
-	    insert_nextexe(cl);
-	}	    
-
-	if (debug_opt) {
-	    struct tm *ftime;
-	    ftime = localtime( &(cl->cl_nextexe) );
-	    debug("  cmd %s next exec %d/%d/%d wday:%d %02d:%02d",
-		  cl->cl_shell, (ftime->tm_mon + 1), ftime->tm_mday,
-		  (ftime->tm_year + 1900), ftime->tm_wday,
-		  ftime->tm_hour, ftime->tm_min); 
-	} 
-	
-
-	cl->cl_next = cf->cf_line_base;
-	cf->cf_line_base = cl;
-	Alloc(cl, CL);
     }
+
+    /* free last cl Alloc : unused */
+    free(cl);
+
     /* check for an error */
     if ( ferror(ff) != 0 )
 	error("file %s is truncated : you should reinstall it with fcrontab",
 	      file_name);
 
-    /* free last calloc : unused */
-    free(cl);
-    
     fclose(ff);
 
-    return 0;
+    return OK;
+
+  err:
+    if ( ff != NULL)
+	fclose(ff);
+
+    if ( cl != NULL && cl->cl_next == NULL ) {
+	/* line is not yet in the line list of the file : free it */
+	if ( cl->cl_shell ) free(cl->cl_shell);
+	if ( cl->cl_runas) free(cl->cl_runas);
+	if ( cl->cl_mailto) free(cl->cl_mailto);
+	free(cl);
+    }
+
+    /* check if we have started to read the lines and env var */
+    if ( cl != NULL ) { 
+	/* insert the line in the file list in order to be able to free
+	 * the memory using delete_file() */
+    
+	cf->cf_next = file_base;
+	file_base = cf;
+
+	delete_file(cf->cf_user);
+	
+    }
+    else if (cf->cf_user != NULL)
+	free(cf->cf_user);
+
+    return ERR;
 
 }
 
 
+int
+add_line_to_file(cl_t *cl, cf_t *cf, uid_t runas, char *runas_str, time_t t_save)
+    /* check if the line is valid, and if yes, add it to the file cf */
+{
+    time_t slept = now - t_save;
+
+    if ( cl->cl_shell == NULL || cl->cl_runas == NULL ||
+	 cl->cl_mailto == NULL ) {
+	error("Line is not valid (empty shell, runas or mailto field)"
+	      " : ignored");
+	bzero(cl, sizeof(cl));
+	if (cl->cl_shell) free(cl->cl_shell);
+	if (cl->cl_runas) free(cl->cl_runas);
+	if (cl->cl_mailto) free(cl->cl_mailto);
+	return 1;
+    }
+
+    /* set runas field if necessary (to improve security) */
+    if (runas != ROOTUID) {
+	if (strcmp(cl->cl_runas, runas_str) != 0)
+	    warn("warning: runas(%s) is not owner (%s): overridden.",
+		 cl->cl_runas, runas_str);
+	Set(cl->cl_runas, runas_str);
+    }
+	    
+    /* we need that here because the user's name contained in the
+     * struct cf_t may be required */
+    cl->cl_file = cf;
+
+    /* check if the mailto field is valid */
+    if ( cl->cl_mailto && (*(cl->cl_mailto) == '-' ||
+			   strcspn(cl->cl_mailto, " \t\n") != strlen(cl->cl_mailto) ) ) {
+	error("mailto field \'%s\' is not valid : set to owner %s.", cl->cl_mailto,
+	      cl->cl_file->cf_user);
+	free(cl->cl_mailto);
+	cl->cl_mailto = strdup2(cl->cl_file->cf_user);
+    }
+
+    /* check if the job hasn't been stopped during execution and insert
+     * it in lavg or serial queue if it was in one at fcron's stops  */
+    if (cl->cl_numexe > 0) {
+	cl->cl_numexe = 0;
+	if ( is_lavg(cl->cl_option) ) {
+	    if ( ! is_strict(cl->cl_option) )
+		add_lavg_job(cl);
+	}
+	else if ( is_serial(cl->cl_option) 
+		  || is_serial_once(cl->cl_option) )
+	    add_serial_job(cl);
+	else {
+	    /* job has been stopped during execution :
+	     * launch it again */
+	    warn("job %s did not finish : running it again.",
+		 cl->cl_shell);
+	    set_serial_once(cl->cl_option);
+	    add_serial_job(cl);
+	}
+    }
+
+    if ( is_td(cl->cl_option) ) {
+    
+	/* set the time and date of the next execution  */
+	if ( cl->cl_nextexe <= now ) {
+	    if ( cl->cl_nextexe == 0 )
+		/* the is a line from a new file */
+		set_next_exe(cl, NO_GOTO);		    
+	    else if (cl->cl_runfreq == 1 &&
+		     is_notice_notrun(cl->cl_option))
+		set_next_exe_notrun(cl, SYSDOWN);
+	    else if ( is_bootrun(cl->cl_option) && t_save != 0 
+		      && cl->cl_runfreq != 1) {
+		if ( cl->cl_remain > 0 && --cl->cl_remain > 0 ) {
+		    debug("    cl_remain: %d", cl->cl_remain);
+		}
+		else {
+		    /* run bootrun jobs */
+		    cl->cl_remain = cl->cl_runfreq;
+		    debug("   boot-run %s", cl->cl_shell);
+		    if ( ! is_lavg(cl->cl_option) ) {
+			set_serial_once(cl->cl_option);
+			add_serial_job(cl);
+		    }
+		    else
+			add_lavg_job(cl);			    
+		}
+		set_next_exe(cl, STD);
+	    }
+	    else {
+		if ( is_notice_notrun(cl->cl_option) ) {
+		    /* set next exe and mail user */
+		    struct tm *since2 = localtime(&cl->cl_nextexe);
+		    struct tm since;
+		    memcpy(&since, since2, sizeof(since));
+		    set_next_exe(cl, NO_GOTO);
+		    mail_notrun(cl, SYSDOWN, &since);
+		} 
+		else
+		    set_next_exe(cl, NO_GOTO);
+	    }
+	}
+	else
+	    /* value of nextexe is valid : just insert line in queue */
+	    insert_nextexe(cl);
+    } else {  /* is_td(cl->cl_option) */
+	/* standard @-lines */
+	if ( is_volatile(cl->cl_option) ) {
+	    /* cl_first is always saved for a volatile line */
+ 	    cl->cl_nextexe = now + cl->cl_first;
+ 	} else
+ 	    cl->cl_nextexe += slept;
+ 
+	if ( cl->cl_timefreq < 10 ) {
+	    error("Invalid timefreq for %s: set to 1 day",
+		  cl->cl_shell);
+	    cl->cl_timefreq = 3600*24;
+	}
+	insert_nextexe(cl);
+    }	    
+
+    if (debug_opt) {
+	struct tm *ftime;
+	ftime = localtime( &(cl->cl_nextexe) );
+	debug("  cmd %s next exec %d/%d/%d wday:%d %02d:%02d:%02d",
+	      cl->cl_shell, (ftime->tm_mon + 1), ftime->tm_mday,
+	      (ftime->tm_year + 1900), ftime->tm_wday,
+	      ftime->tm_hour, ftime->tm_min, ftime->tm_sec); 
+    } 
+
+    /* add the current line to the list, and allocate a new line */
+    if ( (cl->cl_id = next_id++) >= ULONG_MAX - 1)
+	next_id = 0;
+    cl->cl_next = cf->cf_line_base;
+    cf->cf_line_base = cl;
+    return 0;
+}
+
 void
 delete_file(const char *user_name)
     /* free a file if user_name is not null 
      *   otherwise free all files */
 {
-    CF *file;
-    CF *prev_file = NULL;
-    CL *line;
-    CL *cur_line;
+    cf_t *file;
+    cf_t *prev_file = NULL;
+    cl_t *line;
+    cl_t *cur_line;
     env_t *env = NULL;
     env_t *cur_env = NULL;
-    struct job *j = NULL;
-    struct job *prev_j;
+    struct job_t *j = NULL;
+    struct job_t *prev_j;
     int i, k;
-    struct CL **s_a = NULL;
+    struct cl_t **s_a = NULL;
 
     file = file_base;
     while ( file != NULL) {
@@ -664,7 +929,7 @@ delete_file(const char *user_name)
 	for ( i = 0; i < serial_array_size; i++)
 	    if (serial_array[i] != NULL && serial_array[i]->cl_file == file ) {
 		if ( ! s_a )
-		    s_a = calloc(serial_array_size, sizeof(CL *));
+		    s_a = calloc(serial_array_size, sizeof(cl_t *));
 		debug("removing %s from serial queue",
 		      serial_array[i]->cl_shell);
 		serial_num--;
@@ -726,7 +991,7 @@ delete_file(const char *user_name)
     }
 
     if (file == NULL)
-	/* file not in list */
+	/* file not in the file list */
 	return;
     
     /* remove file from file list */
@@ -751,70 +1016,30 @@ delete_file(const char *user_name)
 
 
 void
-save_file(CF *file)
-    /* Store the informations relatives to the executions
-     * of tasks at a defined frequency of system's running time */
+save_file(cf_t *arg_file)
+/* Store the informations relatives to the executions
+ * of tasks at a defined frequency of system's running time */
 {
-    CF *cf = NULL;
-    CL *cl = NULL;
-    FILE *f = NULL;
-    CF *start = NULL;
-    env_t *env = NULL;
+    cf_t *file = NULL;
+    cf_t *start_file = NULL;
 
-    if (file != NULL)
-	start = file;
+    if (arg_file != NULL)
+	start_file = arg_file;
     else
-	start = file_base;
+	start_file = file_base;
     	
     
-    for (cf = start; cf; cf = cf->cf_next) {
+    for (file = start_file; file; file = file->cf_next) {
 
-	debug("Saving %s...", cf->cf_user);
-
-	/* open file for writing */
-	if ( (f = fopen(cf->cf_user, "w")) == NULL )
-	    error_e("save");
+	debug("Saving %s...", file->cf_user);
 
+	/* save the file safely : save it to a temporary name, then rename() it */
 	/* chown the file to root:root : this file should only be read and
 	 * modified by fcron (not fcrontab) */
-	if (fchown(fileno(f), 0, 0) != 0)
-	    error_e("Could not fchown \"%s\"", file->cf_user);
+	save_file_safe(file, file->cf_user, "fcron", ROOTUID, ROOTGID, now);
 
-	/* save file : */
-
-	/* put version of frontab file: it permit to daemon not to load
-	 * a file which he won't understand the syntax, for exemple
-	 * a file using a depreciated format generated by an old fcrontab,
-	 * if the syntax has changed */
-	fprintf(f, "fcrontab-" FILEVERSION "\n");
-
-	/* put the user's name : it is used to check his uid has not changed */
- 	fprintf(f, "%s%c", cf->cf_user, '\0');
-
-	/* put the time & date of saving : this is use for calcutating 
-	 * the system down time */
-	fprintf(f, "%ld", now);
-
-	/*   env variables, */
-	for (env = cf->cf_env_base; env; env = env->e_next)
-	    fprintf(f, "%s%c", env->e_val, '\0');
-	
-	fprintf(f, "%c", '\0');
-
-	/* finally, lines. */
-	for (cl = cf->cf_line_base; cl; cl = cl->cl_next) {
-	    if ( fwrite(cl, sizeof(CL), 1, f) != 1 )
-		error_e("save");
-	    fprintf(f, "%s%c", cl->cl_shell, '\0');
-	    fprintf(f, "%s%c", cl->cl_runas, '\0');
-	    fprintf(f, "%s%c", cl->cl_mailto, '\0');
-	}
-    
-	fclose(f);
-
-	if (file != NULL)
+	if (arg_file != NULL)
 	    /* we have to save only a single file */
 	    break ;
-
     }
 }
