@@ -1,8 +1,7 @@
-
 /*
  * FCRON - periodic command scheduler 
  *
- *  Copyright 2000-2001 Thibault Godouet <fcron@free.fr>
+ *  Copyright 2000-2004 Thibault Godouet <fcron@free.fr>
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -22,25 +21,30 @@
  *  `LICENSE' that comes with the fcron source distribution.
  */
 
- /* $Id: database.c,v 1.50 2001/04/14 13:53:50 thib Exp $ */
+ /* $Id: database.c,v 1.70 2003/12/25 22:50:27 thib Exp thib $ */
 
 #include "fcron.h"
 
+#include "database.h"
+#include "job.h"
+#include "getloadavg.h"
+
 int is_leap_year(int year);
 int get_nb_mdays(int year, int mon);
 void set_wday(struct tm *date);
-void goto_non_matching(CL *line, struct tm *tm, char option);
-void run_normal_job(CL *line);
+void goto_non_matching(cl_t *line, struct tm *tm, char option);
+#define END_OF_INTERVAL 1    /* goto_non_matching() : option */
+void run_normal_job(cl_t *line);
 void run_serial_job(void);
 void run_lavg_job(int i);
-void run_queue_job(CL *line);
-
+void run_queue_job(cl_t *line);
+void resize_exe_array(void);
 
 void
 test_jobs(void)
   /* determine which jobs need to be run, and run them. */
 {
-    struct job *j;
+    struct job_t *j;
 
 /*      // */
     debug("Looking for jobs to execute ...");
@@ -69,7 +73,7 @@ test_jobs(void)
 
 
 void
-run_normal_job(CL *line)
+run_normal_job(cl_t *line)
 {
 
     if (line->cl_numexe <= 0 || 
@@ -78,7 +82,7 @@ run_normal_job(CL *line)
 	run_queue_job(line);
     }
     else {
-	warn("    process already running: %s %s", 
+	warn("    process already running: %s's %s", 
 	     line->cl_file->cf_user,
 	     line->cl_shell
 	    );
@@ -113,8 +117,8 @@ run_serial_job(void)
 /*      debug("running next serial job"); */
 /*      //     */
 
-/*      debug("num: %d running:%d  index:%d", serial_num, serial_running, */
-/*  	  serial_array_index); */
+    debug("num: %d running:%d  index:%d", serial_num, serial_running,
+  	  serial_array_index);
     if ( serial_num != 0 ) {
 	run_queue_job(serial_array[serial_array_index]);
 	serial_array[serial_array_index] = NULL;
@@ -128,8 +132,27 @@ run_serial_job(void)
 }
 		    
 
+void 
+resize_exe_array(void)
+    /* make exe_array bigger */
+{
+    struct exe_t *ptr = NULL;
+    short int old_size = exe_array_size;
+
+    debug("Resizing exe_array");
+    exe_array_size = (exe_array_size + EXE_GROW_SIZE);
+	
+    if ( (ptr = calloc(exe_array_size, sizeof(struct exe_t))) == NULL )
+	die_e("could not calloc exe_array");
+
+    memcpy(ptr, exe_array, (sizeof(struct exe_t) * old_size));
+    free(exe_array);
+    exe_array = ptr;
+}
+
+
 void
-run_queue_job(CL *line)
+run_queue_job(cl_t *line)
     /* run a job */
 {
 
@@ -138,20 +161,9 @@ run_queue_job(CL *line)
 /*      // */
 
     /* append job to the list of executed job */
-    if ( exe_num >= exe_array_size ) {
-	struct exe *ptr = NULL;
-	short int old_size = exe_array_size;
+    if ( exe_num >= exe_array_size )
+	resize_exe_array();
 
-	debug("Resizing exe_array");
-	exe_array_size = (exe_array_size + EXE_GROW_SIZE);
-	
-	if ( (ptr = calloc(exe_array_size, sizeof(struct exe))) == NULL )
-	    die_e("could not calloc exe_array");
-
-	memcpy(ptr, exe_array, (sizeof(struct exe) * old_size));
-	free(exe_array);
-	exe_array = ptr;
-    }
     exe_array[exe_num].e_line = line;
 
     run_job(&exe_array[exe_num++]);
@@ -161,15 +173,15 @@ run_queue_job(CL *line)
 
 
 void
-insert_nextexe(CL *line)
+insert_nextexe(cl_t *line)
     /* insert a job the queue list */
 {
-    struct job *newjob;
+    struct job_t *newjob;
 
     if (queue_base != NULL) {
-	struct job *j;
-	struct job *jprev = NULL;
-	struct job *old_entry = NULL;
+	struct job_t *j;
+	struct job_t *jprev = NULL;
+	struct job_t *old_entry = NULL;
 
 	/* find the job in the list */
 	for (j = queue_base; j != NULL ; j = j->j_next)
@@ -199,7 +211,7 @@ insert_nextexe(CL *line)
 
 	if (old_entry == NULL) {
 	    /* this job wasn't in the queue : we append it */
-	    Alloc(newjob, job);
+	    Alloc(newjob, job_t);
 	    newjob->j_line = line;
 	}
 	else
@@ -216,7 +228,7 @@ insert_nextexe(CL *line)
     }
     else {
 	/* no job in queue */
-	Alloc(newjob, job);
+	Alloc(newjob, job_t);
 	newjob->j_line = line;	    
 	queue_base = newjob;
     }
@@ -224,39 +236,45 @@ insert_nextexe(CL *line)
 }
 
 void
-add_serial_job(CL *line)
+add_serial_job(cl_t *line)
     /* add the next queued job in serial queue */
 {
     short int i;
 
-    /* check if the line is already in the serial queue */
+    /* check if the line is already in the serial queue 
+     * (we consider serial jobs currently running as in the queue) */
     if ( (is_serial_sev(line->cl_option) && line->cl_numexe >= UCHAR_MAX) ||
-	 line->cl_numexe > 0) {
+	 (! is_serial_sev(line->cl_option) && line->cl_numexe > 0) ) {
 	debug("already in serial queue %s", line->cl_shell);
 	return;
     }
 
-/*      debug("inserting in serial queue %s", line->cl_shell); */
+    debug("inserting in serial queue %s", line->cl_shell);
 
     if ( serial_num >= serial_array_size ) {
-	if ( serial_num >= SERIAL_QUEUE_MAX )
-	    /* run next job in the queue before adding the new one */
-	    run_serial_job();
+	if ( serial_num >= serial_queue_max ) {
+  	    error("Could not add job : serial queue is full (%d jobs). "
+		  "Consider using option serialonce, fcron's option -m and/or -q : %s",
+  		 serial_queue_max, line->cl_shell);
+	    if ( is_notice_notrun(line->cl_option) )
+		mail_notrun(line, QUEUE_FULL, NULL);
+	    return;
+	}
 	else {
-	    CL **ptr = NULL;
+	    cl_t **ptr = NULL;
 	    short int old_size = serial_array_size;
 
 	    debug("Resizing serial_array");
 	    serial_array_size = (serial_array_size + SERIAL_GROW_SIZE);
 	
-	    if ( (ptr = calloc(serial_array_size, sizeof(CL *))) == NULL )
+	    if ( (ptr = calloc(serial_array_size, sizeof(cl_t *))) == NULL )
 		die_e("could not calloc serial_array");
 
 	    /* copy lines in order to have the first line at the index 0 */
 	    memcpy(ptr + serial_array_index, serial_array,
-		   (sizeof(CL*) * (old_size - serial_array_index)) );
+		   (sizeof(cl_t*) * (old_size - serial_array_index)) );
 	    memcpy(ptr, serial_array + (old_size - serial_array_index),
-		   (sizeof(CL*) * serial_array_index));
+		   (sizeof(cl_t*) * serial_array_index));
 	    serial_array_index = 0;
 	    free(serial_array);
 	    serial_array = ptr;
@@ -280,15 +298,16 @@ add_serial_job(CL *line)
 
 
 void
-add_lavg_job(CL *line)
+add_lavg_job(cl_t *line)
     /* add the next queued job in lavg queue */
     /* WARNING : must be run before a set_next_exe() to get the strict option
      * working correctly */
 {
 
-    /* check if the line is already in the lavg queue */
-    if ( (is_lavg_sev(line->cl_option) && line->cl_numexe >= UCHAR_MAX)
-	 || line->cl_numexe > 0 ) {
+    /* check if the line is already in the lavg queue
+     * (we consider serial jobs currently running as in the queue) */
+    if ( (is_lavg_sev(line->cl_option) && line->cl_numexe >= UCHAR_MAX) ||
+	 (! is_lavg_sev(line->cl_option) &&  line->cl_numexe > 0 ) ) {
 	debug("already in lavg queue %s", line->cl_shell);
 	return;
     }
@@ -298,27 +317,25 @@ add_lavg_job(CL *line)
 	
     /* append job to the list of lavg job */
     if ( lavg_num >= lavg_array_size ) {
-	if ( lavg_num >= LAVG_QUEUE_MAX ) {
-	    /* run the next lavg job (the oldest one) */
-	    register int i;
-	    int j = 0;
-		
-	    for (i = 1; i < lavg_num; i++)
-		if ( lavg_array[i].l_until < lavg_array[j].l_until )
-		    j = i;
-	    run_lavg_job(j);
+	if ( lavg_num >= lavg_queue_max ) {
+  	    error("Could not add job : lavg queue is full (%d jobs). Consider using "
+		  "options lavgonce, until, strict and/or fcron's option -q.",
+  		 lavg_queue_max, line->cl_shell);
+	    if ( is_notice_notrun(line->cl_option) )
+		mail_notrun(line, QUEUE_FULL, NULL);
+	    return;
 	}
 	else {
-	    struct lavg *ptr = NULL;
+	    struct lavg_t *ptr = NULL;
 	    short int old_size = lavg_array_size;
 		
 	    debug("Resizing lavg_array");
 	    lavg_array_size = (lavg_array_size + LAVG_GROW_SIZE);
 		
-	    if ( (ptr = calloc(lavg_array_size, sizeof(lavg))) == NULL )
+	    if ( (ptr = calloc(lavg_array_size, sizeof(lavg_t))) == NULL )
 		die_e("could not calloc lavg_array");
 		
-	    memcpy(ptr, lavg_array, (sizeof(lavg) * old_size));
+	    memcpy(ptr, lavg_array, (sizeof(lavg_t) * old_size));
 	    free(lavg_array);
 	    lavg_array = ptr;
 	}
@@ -330,9 +347,12 @@ add_lavg_job(CL *line)
     if ( is_strict(line->cl_option) && line->cl_runfreq == 1) {
 	struct tm *ft;
 	struct tm ftime;
-	time_t end_of_cur_int = 0;
+	time_t begin_of_cur_int, end_of_cur_int = 0;
 
-	ft = localtime(&line->cl_nextexe);
+	/* handle timezone differences */
+	begin_of_cur_int = line->cl_nextexe - (line->cl_file->cf_tzdiff*3600);
+
+	ft = localtime( &begin_of_cur_int );
 
 	/* localtime() function seem to return every time the same pointer :
 	   it resets our previous changes, so we need to prevent it
@@ -341,7 +361,7 @@ add_lavg_job(CL *line)
 
 	goto_non_matching(line, &ftime, END_OF_INTERVAL);
 
-	end_of_cur_int = mktime(&ftime);
+	end_of_cur_int = mktime(&ftime) + (line->cl_file->cf_tzdiff * 3600);
 
 	if ((line->cl_until > 0) && (line->cl_until + now < end_of_cur_int))
 	    lavg_array[lavg_num].l_until = line->cl_until + now;
@@ -364,7 +384,7 @@ wait_chld(void)
 {
     short int i = 0;
     int pid;
-    CL *line = NULL;
+    cl_t *line = NULL;
 
 
 /*      // */
@@ -374,7 +394,7 @@ wait_chld(void)
     while ( (pid = wait3(NULL, WNOHANG, NULL)) > 0 ) {
 	i = 0;
 	while ( i < exe_num ) {
-	    if (pid == exe_array[i].e_pid) {
+	    if (pid == exe_array[i].e_ctrl_pid) {
 		if ( exe_array[i].e_line == NULL ) {
 		    /* the corresponding file has been removed from memory */
 		    debug("job finished: pid %d", pid);
@@ -388,12 +408,12 @@ wait_chld(void)
 		    
 		    if ( is_serial_once(line->cl_option) ) {
 			clear_serial_once(line->cl_option);
-			if ( --serial_running <= 0 )
+			if ( --serial_running < serial_max_running )
 			    run_serial_job();
 		    }
 		    else if ( is_serial(line->cl_option)
 			      && ! is_lavg(line->cl_option) ) {
-			if (--serial_running <= 0)
+			if (--serial_running < serial_max_running)
 			    run_serial_job();
 		    }
 		    else if ( is_lavg(line->cl_option) && 
@@ -428,7 +448,7 @@ wait_all(int *counter)
     while ( (*counter > 0) && (pid = wait3(NULL, 0, NULL)) > 0 ) {
 	i = 0;
 	while ( i < exe_num ) {
-	    if (pid == exe_array[i].e_pid) {
+	    if (pid == exe_array[i].e_ctrl_pid) {
 		if ( exe_array[i].e_line == NULL ) {
 		    /* the corresponding file has been removed from memory */
 		    debug("job finished: pid %d", pid);
@@ -499,7 +519,7 @@ set_wday(struct tm *date)
    * and add this number modulo 7 to the wday number */
 {
     long nod = 0;
-    register int i;
+    int i;
 
     /* we add the number of days of each previous years */
     for (i =  (date->tm_year - 1); i >= 100 ; i--)
@@ -526,7 +546,7 @@ set_wday(struct tm *date)
 
 
 void
-goto_non_matching(CL *line, struct tm *ftime, char option)
+goto_non_matching(cl_t *line, struct tm *ftime, char option)
     /* search the first the nearest time and date that does
      * not match the line */
 {
@@ -564,12 +584,16 @@ goto_non_matching(CL *line, struct tm *ftime, char option)
 			} else
 			    ftime->tm_mday += to_add;
 		    } else {
-			ftime->tm_mday = 1;
 			if (is_freq_mons(line->cl_option)) {
 			    if (ftime->tm_mday >= 15)
 				ftime->tm_mon++;
 			    ftime->tm_mday = 15;
 			}
+			else
+			    /* weird : we have the bit freq_mid set, but
+			     * none of freq_{mins|hour|days|dow|mons} is set :
+			     * we does nothing */
+			    ;
 		    }
 		}
 	    }
@@ -626,10 +650,10 @@ goto_non_matching(CL *line, struct tm *ftime, char option)
 	if (option != END_OF_INTERVAL) {
 	    if (debug_opt)
 		set_wday(ftime);
-	    debug("   %s beginning of next period %d/%d/%d wday:%d %02d:%02d",
-		  line->cl_shell, (ftime->tm_mon + 1), ftime->tm_mday,
-		  (ftime->tm_year + 1900), ftime->tm_wday,
-		  ftime->tm_hour, ftime->tm_min);
+	    debug("   %s beginning of next period %d/%d/%d wday:%d %02d:%02d "
+		  "(tzdiff=%d)", line->cl_shell, (ftime->tm_mon + 1),
+		  ftime->tm_mday, (ftime->tm_year + 1900), ftime->tm_wday,
+		  ftime->tm_hour, ftime->tm_min, line->cl_file->cf_tzdiff);
 
 	    return;
 	}
@@ -750,32 +774,35 @@ goto_non_matching(CL *line, struct tm *ftime, char option)
 	    }
 	}
 	
-	debug("   %s %s %d/%d/%d wday:%d %02d:%02d", line->cl_shell,
+	debug("   %s %s %d/%d/%d wday:%d %02d:%02d (tzdiff=%d)",line->cl_shell,
 	      (option == STD) ? "first non matching" : "end of interval",
-	      (ftime->tm_mon + 1), ftime->tm_mday,
-	      (ftime->tm_year + 1900), ftime->tm_wday,
-	      ftime->tm_hour, ftime->tm_min);
+	      (ftime->tm_mon + 1), ftime->tm_mday, (ftime->tm_year + 1900),
+	      ftime->tm_wday, ftime->tm_hour, ftime->tm_min,
+	      line->cl_file->cf_tzdiff);
 	return;
     }
 }
 
 
 void 
-set_next_exe(CL *line, char option)
-  /* set the cl_nextexe of a given CL and insert it in the queue */
+set_next_exe(cl_t *line, char option)
+  /* set the cl_nextexe of a given cl_t and insert it in the queue */
 {
 
     if ( is_td(line->cl_option) ) {
 
 	struct tm *ft;
 	struct tm ftime;
-	register int i;
+	time_t nextexe = 0;
+	int i;
 	int max;
-	register char has_changed = 0;
+	char has_changed = 0;
 	/* to prevent from invinite loop with unvalid lines : */
 	short int year_limit = MAXYEAR_SCHEDULE_TIME;
+	/* timezone difference */
+	time_t now_tz = now - (line->cl_file->cf_tzdiff * 3600);
 
-	ft = localtime(&now);
+	ft = localtime(&now_tz);
 
 	/* localtime() function seem to return every time the same pointer :
 	   it resets our previous changes, so we need to prevent it
@@ -784,10 +811,14 @@ set_next_exe(CL *line, char option)
 
 	ftime.tm_isdst = -1;
 
-	/* to prevent multiple execution in the same minute */
-	ftime.tm_min += 1;
-	ftime.tm_sec = 0;
-	if (line->cl_runfreq==1 && option != NO_GOTO && option != NO_GOTO_LOG)
+	/* to prevent multiple execution of &-jobs in the same minute
+ 	 * (but not if the user has explicitely asked to run jobs immediately) */
+ 	if (first_sleep > 0 || option == STD || line->cl_runfreq != 1) {
+	    ftime.tm_min += 1;
+	    ftime.tm_sec = 0;
+	}
+	    
+	if (line->cl_runfreq == 1 && option != NO_GOTO && option != NO_GOTO_LOG)
 	    goto_non_matching(line, &ftime, STD);
 
       setMonth:
@@ -858,7 +889,7 @@ set_next_exe(CL *line, char option)
 		goto setDay;
 	    }
 	} else {  /* dayor */
-	    register int j;
+	    int j;
 
 	    set_wday(&ftime);
 
@@ -918,14 +949,41 @@ set_next_exe(CL *line, char option)
 	}
 	ftime.tm_min = i;
    
-      set_cl_nextexe:    
-	line->cl_nextexe = mktime(&ftime);
+      set_cl_nextexe:
+	/* set cl_nextexe (handle the timezone differences) */
+	nextexe = mktime(&ftime);
 
-	if ( option != NO_GOTO )
-	    debug("   cmd: %s next exec %d/%d/%d wday:%d %02d:%02d",
-		  line->cl_shell, (ftime.tm_mon + 1), ftime.tm_mday,
-		  (ftime.tm_year + 1900), ftime.tm_wday,
-		  ftime.tm_hour, ftime.tm_min);
+	if ( is_random(line->cl_option) ) {
+	    /* run the job at a random time during its interval of execution */
+	    struct tm intend;
+	    time_t intend_int;
+
+	    debug("   cmd: %s begin int exec %d/%d/%d wday:%d %02d:%02d "
+		  "(tzdiff=%d)", line->cl_shell, (ftime.tm_mon + 1),
+		  ftime.tm_mday, (ftime.tm_year + 1900), ftime.tm_wday,
+		  ftime.tm_hour, ftime.tm_min, line->cl_file->cf_tzdiff);
+
+	    memcpy(&intend, &ftime, sizeof(intend));
+	    goto_non_matching(line, &intend, END_OF_INTERVAL);
+	    intend_int = mktime(&intend);
+
+	    /* set a random time to add to the first allowed time of execution */
+	    nextexe += ( (i= intend_int - nextexe) > 0) ? 
+		(time_t)(((float)i * (float)rand())/(float)RAND_MAX) : 0;
+	}
+
+	line->cl_nextexe = nextexe + (line->cl_file->cf_tzdiff * 3600);
+
+	if ( option != NO_GOTO ) {
+	    if ( is_random(line->cl_option) ) {
+		ft = localtime(&nextexe);
+		memcpy(&ftime, ft, sizeof(ftime));
+	    }
+	    debug("   cmd: %s next exec %d/%d/%d wday:%d %02d:%02d "
+		  "(tzdiff=%d)", line->cl_shell, (ftime.tm_mon + 1),
+		  ftime.tm_mday, (ftime.tm_year + 1900), ftime.tm_wday,
+		  ftime.tm_hour, ftime.tm_min, line->cl_file->cf_tzdiff);
+	}
 	
     }
     else {
@@ -941,11 +999,11 @@ set_next_exe(CL *line, char option)
 
 
 void
-set_next_exe_notrun(CL *line, char context)
+set_next_exe_notrun(cl_t *line, char context)
     /* set the time of the next execution and send a mail to tell user his job
      * has not run if necessary */
 {
-    time_t next_period = 0;
+    time_t previous_period = 0, next_period = 0;
     struct tm *ft = NULL;
     struct tm ftime, last_nextexe;
     char set_next_exe_opt = 0;
@@ -955,14 +1013,16 @@ set_next_exe_notrun(CL *line, char context)
 /*  // */
 
     if (context == SYSDOWN) {
-	ft = localtime(&line->cl_nextexe);
+	/* handle timezone differences */
+	previous_period = line->cl_nextexe - (line->cl_file->cf_tzdiff * 3600);
 	set_next_exe_opt = NO_GOTO;
     }
     else {
-	ft = localtime(&now);
+	previous_period = now - (line->cl_file->cf_tzdiff * 3600);
 	set_next_exe_opt = NO_GOTO_LOG;
     }
-    
+    ft = localtime(&previous_period);
+
     /* localtime() function seem to return every time the same pointer :
        it resets our previous changes, so we need to prevent it
        ( localtime() is used in the debug() function) */
@@ -972,7 +1032,7 @@ set_next_exe_notrun(CL *line, char context)
     
     ftime.tm_sec = 0;
     goto_non_matching(line, &ftime, STD);
-    next_period = mktime(&ftime);
+    next_period = mktime(&ftime) + (line->cl_file->cf_tzdiff * 3600);
 
     set_next_exe(line, set_next_exe_opt);
     if ( line->cl_nextexe >= next_period ) {
@@ -983,13 +1043,12 @@ set_next_exe_notrun(CL *line, char context)
 }
 
 void
-mail_notrun(CL *line, char context, struct tm *since)
-    /* send a mail to tell user a %-job has not run (and why) */
+mail_notrun(cl_t *line, char context, struct tm *since)
+    /* send a mail to tell user a job has not run (and why) */
 {
     int pid = 0;
-    int fd = 0;
-    struct tm *next2 = NULL, next;
-    char buf[LINE_LEN];
+    FILE *mailf = 0;
+    struct tm *time2 = NULL, time;
 
     switch ( pid = fork() ) {
     case -1:
@@ -1006,67 +1065,64 @@ mail_notrun(CL *line, char context, struct tm *since)
 	      line->cl_shell, pid);
 /*  // */
 
-	/* create a entry in exe_array */
-	if ( exe_num >= exe_array_size ) {
-	    struct exe *ptr = NULL;
-	    short int old_size = exe_array_size;
-	    
-	    debug("Resizing exe_array");
-	    exe_array_size = (exe_array_size + EXE_GROW_SIZE);
-	    
-	    if ( (ptr = calloc(exe_array_size, sizeof(struct exe))) == NULL )
-		die_e("could not calloc exe_array");
-	    
-	    memcpy(ptr, exe_array, (sizeof(struct exe) * old_size));
-	    free(exe_array);
-	    exe_array = ptr;
-	}
+	/* create an entry in exe_array */
+	if ( exe_num >= exe_array_size )
+	    resize_exe_array();
 	/* set line to NULL as this is not a line ... */
 	exe_array[exe_num].e_line = NULL;
-	exe_array[exe_num].e_pid = pid;
+	exe_array[exe_num].e_ctrl_pid = pid;
 	exe_num++;
 	return;
     }
 
-    next2 = localtime(&line->cl_nextexe);
-    memcpy(&next, next2, sizeof(next));
+    if ( context == QUEUE_FULL )
+	time2 = localtime(&now);
+    else
+	time2 = localtime(&line->cl_nextexe);
+    memcpy(&time, time2, sizeof(time));
 
     /* create a temp file, and write in it the message to send */
-    fd = create_mail(line, "Non-execution of fcron job");
+    mailf = create_mail(line, "Non-execution of fcron job");
 
-    if (context == SYSDOWN) {
-	snprintf(buf, sizeof(buf), "Line %s has not run since and including "
+    switch ( context ) {
+    case SYSDOWN:
+	fprintf(mailf, "Line %s has not run since and including "
 		 "%d/%d/%d wday:%d %02d:%02d\ndue to system's down state.\n",
 		 line->cl_shell, (since->tm_mon + 1), since->tm_mday,
 		 (since->tm_year + 1900), since->tm_wday, since->tm_hour,
 		 since->tm_min);
-	xwrite(fd, buf);
-	snprintf(buf,sizeof(buf),"It will be next executed at %d/%d/%d wday:"
-		 "%d %02d:%02d\n", (next.tm_mon + 1), next.tm_mday,
-		 (next.tm_year+1900), next.tm_wday, next.tm_hour, next.tm_min);
-	xwrite(fd, buf);
-    }
-    else if (context == LAVG) {
-	snprintf(buf, sizeof(buf), "Line %s has not run since and including "
+	fprintf(mailf, "It will be next executed at %d/%d/%d wday:"
+		 "%d %02d:%02d\n", (time.tm_mon + 1), time.tm_mday,
+		 (time.tm_year+1900), time.tm_wday, time.tm_hour, time.tm_min);
+	break;
+    case LAVG:
+	fprintf(mailf, "Line %s has not run since and including "
 		 "%d/%d/%d wday:%d %02d:%02d\n", line->cl_shell,
 		 (since->tm_mon + 1), since->tm_mday, (since->tm_year + 1900),
 		 since->tm_wday, since->tm_hour, since->tm_min);
-	xwrite(fd, buf);
-	snprintf(buf, sizeof(buf), "due to a too high system load average or "
+	fprintf(mailf, "due to a too high system load average or "
 		 "too many lavg-serial jobs.\n");
-	xwrite(fd, buf);	
-	snprintf(buf, sizeof(buf), "It will be next executed at %d/%d/%d "
-		 "wday:%d %02d:%02d\n", (next.tm_mon + 1), next.tm_mday,
-		 (next.tm_year+1900), next.tm_wday, next.tm_hour, next.tm_min);
-	xwrite(fd, buf);
+	fprintf(mailf, "It will be next executed at %d/%d/%d "
+		 "wday:%d %02d:%02d\n", (time.tm_mon + 1), time.tm_mday,
+		 (time.tm_year+1900), time.tm_wday, time.tm_hour, time.tm_min);
+	break;
+    case QUEUE_FULL:
+	fprintf(mailf,"Line %s couldn't be added to lavg or serial queue which"
+		 " is full ( %d/%d/%d wday:%d %02d:%02d ).\n", line->cl_shell,
+		 (time.tm_mon + 1), time.tm_mday, (time.tm_year + 1900),
+		 time.tm_wday, time.tm_hour, time.tm_min);
+	fprintf(mailf, "Consider using options lavgonce, until, strict, "
+		 "serialonce and/or fcron's option -m.\n");
+	fprintf(mailf, "Note that job %s has not run.\n", line->cl_shell);
+	break;
     }
     
     /* become user (for security reasons) */
-    if (change_user(line->cl_runas) < 0)
+    if (change_user(line) < 0)
 	return ;
 
     /* then, send mail */
-    launch_mailer(line, fd);
+    launch_mailer(line, mailf);
     
     /* we should not come here : launch_mailer does not return */
     error("mail_notrun : launch_mailer failed");
@@ -1086,12 +1142,8 @@ check_lavg(time_t lim)
     return tts;
 #else
 
-    register int i = 0;
-    double l_avg[3];
-
-    l_avg[0] = 0;
-    l_avg[1] = 0;
-    l_avg[2] = 0;
+    int i = 0;
+    double l_avg[3]= {0, 0, 0};
 
     /* first, check if some lines must be executed because of until */
     while ( i < lavg_num )
@@ -1104,7 +1156,8 @@ check_lavg(time_t lim)
 			    lavg_array[i].l_line->cl_shell);
 
 		/* set time of the next execution and send a mail if needed */
-		if ( is_notice_notrun(lavg_array[i].l_line->cl_option) )
+		if ( is_td(lavg_array[i].l_line->cl_option) &&
+		     is_notice_notrun(lavg_array[i].l_line->cl_option) )
   		    set_next_exe_notrun(lavg_array[i].l_line, LAVG);
 		else
 		    set_next_exe(lavg_array[i].l_line, NO_GOTO_LOG);
@@ -1143,7 +1196,7 @@ check_lavg(time_t lim)
     i = 0;
     while ( i < lavg_num ) {
 	/* check if the line should be executed */
-	if ( lavg_serial_running > 0 && 
+	if ( lavg_serial_running >= serial_max_running && 
 	     is_serial(lavg_array[i].l_line->cl_option) ) {
 	    i++;
 	    continue;
@@ -1202,7 +1255,7 @@ time_to_sleep(time_t lim)
     if ( (tts = tts - ti) < 0)
 	tts = 0;
 
-/*      // debug("Time to sleep: %lds", tts); */
+/*      debug("Time to sleep: %lds", tts); */
 
     return tts;
 }
