@@ -2,7 +2,7 @@
 /*
  * FCRON - periodic command scheduler 
  *
- *  Copyright 2000-2001 Thibault Godouet <fcron@free.fr>
+ *  Copyright 2000-2004 Thibault Godouet <fcron@free.fr>
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -22,7 +22,7 @@
  *  `LICENSE' that comes with the fcron source distribution.
  */
 
- /* $Id: fcrontab.c,v 1.33 2001/04/14 14:04:23 thib Exp $ */
+ /* $Id: fcrontab.c,v 1.64 2003/12/25 22:42:45 thib Exp thib $ */
 
 /* 
  * The goal of this program is simple : giving a user interface to fcron
@@ -42,14 +42,20 @@
 
 #include "fcrontab.h"
 
-char rcs_info[] = "$Id: fcrontab.c,v 1.33 2001/04/14 14:04:23 thib Exp $";
+#include "allow.h"
+#include "fileconf.h"
+#include "temp_file.h"
+#include "read_string.h"
+
+char rcs_info[] = "$Id: fcrontab.c,v 1.64 2003/12/25 22:42:45 thib Exp thib $";
 
 void info(void);
 void usage(void);
-void sig_daemon(void);
-pid_t read_pid(void);
 
 
+/* used in temp_file() */
+char *tmp_path = "/tmp/";
+
 /* command line options */
 char rm_opt = 0;
 char list_opt = 0;
@@ -65,25 +71,32 @@ char debug_opt = 0;       /* set to 1 if we are in debug mode */
 #endif
 
 char *user = NULL;
+char *runas = NULL;
 uid_t uid = 0;
 uid_t asuid = 0;
 gid_t asgid = 0;
 uid_t fcrontab_uid = 0;
 gid_t fcrontab_gid = 0;
-char  *cdir = FCRONTABS;
 
 char need_sig = 0;           /* do we need to signal fcron daemon */
 
-char *orig_dir = NULL;
-CF *file_base = NULL;
-char buf[FNAME_LEN];
-char file[FNAME_LEN];
+char orig_dir[PATH_LEN];
+cf_t *file_base = NULL;
+char buf[PATH_LEN];
+char file[PATH_LEN];
 
 /* needed by log part : */
 char *prog_name = NULL;
 char foreground = 1;
+char dosyslog = 1;
 pid_t daemon_pid = 0;
 
+#ifdef HAVE_LIBPAM
+int conv_pam(int num_msg, const struct pam_message **msgm,
+	     struct pam_response **response, void *appdata_ptr);
+pam_handle_t *pamh = NULL;
+const struct pam_conv apamconv = { conv_pam, NULL };
+#endif /* HAVE_LIBPAM */
 
 void
 info(void)
@@ -92,7 +105,7 @@ info(void)
 {
     fprintf(stderr,
 	    "fcrontab " VERSION_QUOTED " - user interface to daemon fcron\n"
-	    "Copyright 2000-2001 Thibault Godouet <fcron@free.fr>\n"
+	    "Copyright 2000-2004 Thibault Godouet <fcron@free.fr>\n"
 	    "This program is free software distributed WITHOUT ANY WARRANTY.\n"
             "See the GNU General Public License for more details.\n"
 	);
@@ -107,8 +120,8 @@ usage(void)
   /*  print a help message about command line options and exit */
 {
     fprintf(stderr, 
-	    "fcrontab [-u user] [-n] file\n"
-	    "fcrontab [-u user] { -l | -r | -e | -z } [-n]\n"
+	    "fcrontab [-n] file [user|-u user]\n"
+	    "fcrontab { -l | -r | -e | -z } [-n] [user|-u user]\n"
 	    "fcrontab -h\n"
 	    "  -u user    specify user name.\n"
 	    "  -l         list user's current fcrontab.\n"
@@ -116,8 +129,10 @@ usage(void)
 	    "  -e         edit user's current fcrontab.\n"
 	    "  -z         reinstall user's fcrontab from source code.\n"
 	    "  -n         ignore previous version of file.\n"
+	    "  -c f       make fcrontab use config file f.\n"
 	    "  -d         set up debug mode.\n"
 	    "  -h         display this help message.\n"
+	    "  -V         display version & infos about fcrontab.\n"
 	    "\n"
 	);
     
@@ -125,128 +140,39 @@ usage(void)
 }
 
 
-pid_t
-read_pid(void)
-    /* return fcron daemon's pid if running.
-     * otherwise return 0 */
-{
-    FILE *fp = NULL;
-    pid_t pid = 0;
-    
-    if ((fp = fopen(PIDFILE, "r")) != NULL) {
-	fscanf(fp, "%d", &pid);    
-	fclose(fp);
-    }
-
-    return pid;
-}
-
-
-void
-sig_daemon(void)
-    /* send SIGHUP to daemon to tell him configuration has changed */
-    /* SIGHUP is sent once 10s before the next minute to avoid
-     * some bad users to block daemon by sending it SIGHUP all the time */
-{
-    /* we don't need to make root wait */
-    if (uid != 0) {
-	time_t t = 0;
-	int sl = 0;
-	FILE *fp = NULL;
-	int	fd = 0;
-	struct tm *tm = NULL;
-
-
-	t = time(NULL);
-	tm = localtime(&t);
-    
-	if ( (sl = 60 - (t % 60) - 10) < 0 ) {
-	    if ( (tm->tm_min = tm->tm_min + 2) >= 60 ) {
-		tm->tm_hour++;
-		tm->tm_min -= 60;
-	    }
-	    snprintf(buf, sizeof(buf), "%02dh%02d", tm->tm_hour, tm->tm_min);
-	    sl = 60 - (t % 60) + 50;
-	} else {
-	    if ( ++tm->tm_min >= 60 ) {
-		tm->tm_hour++;
-		tm->tm_min -= 60;
-	    }
-	    snprintf(buf, sizeof(buf), "%02dh%02d", tm->tm_hour, tm->tm_min);
-	}
-	fprintf(stderr, "Modifications will be taken into account"
-		" at %s.\n", buf);
-
-
-#if defined(HAVE_SETREGID) && defined(HAVE_SETREUID)
-	if (seteuid(fcrontab_uid) != 0)
-	    die_e("seteuid(fcrontab_uid[%d])", fcrontab_uid);
-#endif
-
-	/* try to create a lock file */
-	if ((fd = open(FCRONTABS "/fcrontab.sig", O_RDWR|O_CREAT, 0644)) == -1
-	    || ((fp = fdopen(fd, "r+")) == NULL) )
-	    die_e("can't open or create " FCRONTABS "/fcrontab.sig");	
-    
-	if ( flock(fd, LOCK_EX|LOCK_NB) != 0 ) {
-	    debug("fcrontab is already waiting for signalling the daemon :"
-		  " exiting.");
-	    return;
-	}
-
-
-	(void) fcntl(fd, F_SETFD, 1);
-
-	/* abandon fd and fp even though the file is open. we need to
-	 * keep it open and locked, but we don't need the handles elsewhere.
-	 */
-
-	switch ( fork() ) {
-	case -1:
-	    remove(FCRONTABS "/fcrontab.sig");
-	    die_e("could not fork : daemon as not been signaled");
-	    break;
-	case 0:
-	    /* child */
-	    break;
-	default:
-	    /* parent */
-	    return;
-	}
-
-	sleep(sl);
-    
-	remove(FCRONTABS "/fcrontab.sig");
-    }
-    else
-	fprintf(stderr, "Modifications will be taken into account"
-		" right now.\n");
-
-    if ( (daemon_pid = read_pid()) == 0 )
-	/* daemon is not running any longer : we exit */
-	return ;
-
-    if ( kill(daemon_pid, SIGHUP) != 0)
-	die_e("could not send SIGHUP to daemon (pid %d)", daemon_pid);
-
-}
-
-
-
 void
 xexit(int exit_val)
     /* launch signal if needed and exit */
 {
+    pid_t pid = 0;
+
     if ( need_sig == 1 ) {
-	/* check if daemon is running */
-	if ( (daemon_pid = read_pid()) != 0 )
-	    /* warning : we change euid to fcrontab_iud in sig_daemon() */
-	    sig_daemon();
-	else
-	    fprintf(stderr, "fcron is not running :\n  modifications will"
-		    " be taken into account at its next execution.\n");
+	
+	/* fork and exec fcronsighup */
+	switch ( pid = fork() ) {
+	case 0:
+	    /* child */
+	    execl(BINDIREX "/fcronsighup", BINDIREX "/fcronsighup",
+		  fcronconf, NULL);
+	    die_e("Could not exec " BINDIREX " fcronsighup");
+	    break;
+
+	case -1:
+	    die_e("Could not fork (fcron has not been signaled)");
+	    break;
+
+	default:
+	    /* parent */
+	    waitpid(pid, NULL, 0);
+	    break;
+	}
     }
 
+#ifdef HAVE_LIBPAM
+    pam_setcred(pamh, PAM_DELETE_CRED | PAM_SILENT);
+    pam_end(pamh, pam_close_session(pamh, PAM_SILENT));
+#endif
+
     exit(exit_val);
 
 }
@@ -256,43 +182,54 @@ int
 copy(char *orig, char *dest)
     /* copy orig file to dest */
 {
-    FILE *from = NULL, *to = NULL;
-    char c;
+    int from;
+    int to_fd;
+    int nb;
+    char *copy_buf[LINE_LEN];
 
-    if ( (from = fopen(orig, "r")) == NULL) {
-	error_e("copy: orig");
+    if ( (from = open(orig, O_RDONLY)) == -1) {
+	error_e("copy: open(orig)");
 	return ERR;
     }
     /* create it as fcrontab_uid (to avoid problem if user's uid changed)
      * except for root. Root requires filesystem uid root for security
      * reasons */
-#if defined(HAVE_SETREGID) && defined(HAVE_SETREUID)
-    if (asuid != 0 && seteuid(fcrontab_uid) != 0)
-	error_e("seteuid(fcrontab_uid[%d])", fcrontab_uid);
+#ifdef USE_SETE_ID
+    if (asuid == ROOTUID) {
+	if (seteuid(ROOTUID) != 0)
+	    error_e("seteuid(ROOTUID)");
+    } 
+    else {
+    	if (seteuid(fcrontab_uid) != 0)
+	    error_e("seteuid(fcrontab_uid[%d])", fcrontab_uid);
+    }
 #endif
-    if ((to = fopen(dest, "w")) == NULL) {
+    to_fd = open(dest, O_WRONLY | O_CREAT | O_TRUNC | O_SYNC, S_IRUSR|S_IWUSR|S_IRGRP);
+    if (to_fd == -1) {
 	error_e("copy: dest");
 	return ERR;
     }
-#if defined(HAVE_SETREGID) && defined(HAVE_SETREUID)
-    if (asuid != 0 && seteuid(uid) != 0)
+#ifdef USE_SETE_ID
+    if (asuid != ROOTUID && seteuid(uid) != 0)
 	die_e("seteuid(uid[%d])", uid);
 #endif
-    if (asuid == 0 ) {
-	if ( fchmod(fileno(to), S_IWUSR | S_IRUSR) != 0 )
+    if (asuid == ROOTUID ) {
+	if ( fchmod(to_fd, S_IWUSR | S_IRUSR) != 0 )
 	    error_e("Could not fchmod %s to 600", dest);
-	if ( fchown(fileno(to), 0, fcrontab_gid) != 0 )
+	if ( fchown(to_fd, ROOTUID, fcrontab_gid) != 0 )
 	    error_e("Could not fchown %s to root", dest);
     }
 
-    while ( (c = getc(from)) != EOF )
-	if ( putc(c, to) == EOF ) {
+    while ( (nb = read(from, copy_buf, sizeof(copy_buf))) != -1 && nb != 0 )
+	if ( write(to_fd, copy_buf, nb) != nb ) {
 	    error("Error while copying file. Aborting.\n");
+	    close(from);
+	    close(to_fd);
 	    return ERR;
 	}
 
-    fclose(from);
-    fclose(to);
+    close(from);
+    close(to_fd);
     
     return OK;
 }
@@ -304,7 +241,7 @@ remove_fcrontab(char rm_orig)
     /* note : the binary fcrontab is removed by fcron */
 {
     int return_val = OK;
-    FILE *f;
+    int fd;
 
     if ( rm_orig )
 	explain("removing %s's fcrontab", user);
@@ -317,7 +254,7 @@ remove_fcrontab(char rm_orig)
 	    error_e("could not remove %s", buf);		
     }
 
-#if defined(HAVE_SETREGID) && defined(HAVE_SETREUID)
+#ifdef USE_SETE_ID
     if (seteuid(fcrontab_uid) != 0)
 	error_e("seteuid(fcrontab_uid[%d])", fcrontab_uid);
 #endif
@@ -330,16 +267,23 @@ remove_fcrontab(char rm_orig)
     /* finally create a file in order to tell the daemon
      * a file was removed, and launch a signal to daemon */
     snprintf(buf, sizeof(buf), "rm.%s", user);
-    f = fopen(buf, "w");
-    fclose(f);
-    
-    need_sig = 1;
+    fd = open(buf, O_CREAT | O_TRUNC | O_EXCL, S_IRUSR | S_IWUSR);
 
-#if defined(HAVE_SETREGID) && defined(HAVE_SETREUID)
+#ifdef USE_SETE_ID
     if (seteuid(uid) != 0)
 	die_e("seteuid(uid[%d])", uid);
 #endif
 
+    if ( fd == -1 ) {
+	if ( errno != EEXIST )
+	    error_e("Can't create file %s", buf);
+    }
+    else if ( asuid == ROOTUID && fchown(fd, ROOTUID, fcrontab_gid) != 0 )
+	error_e("Could not fchown %s to root", buf);
+    close(fd);
+    
+    need_sig = 1;
+
     return return_val;
 
 }
@@ -358,16 +302,17 @@ write_file(char *file)
 
     if ( file_base->cf_line_base == NULL ) {
 	/* no entries */
-	explain("%s's fcrontab contains no entries", user);
+	explain("%s's fcrontab contains no entries : removed.", user);
 	remove_fcrontab(0);
     } 
     else {
-	/* write that list in a temp file on disk */
+	/* write the binary fcrontab on disk */
 	snprintf(buf, sizeof(buf), "new.%s", user);
-	save_file(buf);
+	if ( save_file(buf) != OK )
+	    return_val = ERR;
     }
 
-    /* copy original file to FCRONTABS dir */
+    /* copy original file to fcrontabs dir */
     snprintf(buf, sizeof(buf), "%s.orig", user);
     if ( copy(file, buf) == ERR )
 	return_val = ERR;
@@ -378,11 +323,10 @@ write_file(char *file)
 int
 make_file(char *file)
 {
-
     explain("installing file %s for user %s", file, user);
 
     /* read file and create a list in memory */
-    switch ( read_file(file, user) ) {
+    switch ( read_file(file) ) {
     case 2:
     case OK:
 
@@ -409,7 +353,7 @@ void
 list_file(char *file)
 {
     FILE *f = NULL;
-    char c;
+    int c;
 
     explain("listing %s's fcrontab", user);
 
@@ -437,36 +381,33 @@ edit_file(char *buf)
     /* copy file to a temp file, edit that file, and install it
        if necessary */
 {
-    char *editor = NULL;
+    char *cureditor = NULL;
+    char editorcmd[PATH_LEN];
     pid_t pid;
     int status;
     struct stat st;
     time_t mtime = 0;
-    char tmp[FNAME_LEN];
+    char *tmp_str;
     FILE *f, *fi;
     int file = 0;
-    char c;
+    int c;
     char correction = 0;
+    short return_val = EXIT_OK;
 
-    explain("fcrontabs : editing %s's fcrontab", user);	
+    explain("fcrontab : editing %s's fcrontab", user);	
 
-    if ( (editor = getenv("VISUAL")) == NULL || strcmp(editor, "\0") == 0 )
-	if( (editor = getenv("EDITOR")) == NULL || strcmp(editor, "\0") == 0 )
-	    editor = EDITOR;
+    if ((cureditor=getenv("VISUAL")) == NULL || strcmp(cureditor, "\0") == 0 )
+	if((cureditor=getenv("EDITOR"))==NULL || strcmp(cureditor, "\0") == 0 )
+	    cureditor = editor;
 	
-    sprintf(tmp, "/tmp/fcrontab.%d", getpid());
-
-    if ( (file = open(tmp, O_CREAT|O_EXCL|O_WRONLY, 0600)) == -1 ) {
-	error_e("could not create file %s", tmp);
-	goto exiterr;
-    }
+    file = temp_file(&tmp_str);
     if ( (fi = fdopen(file, "w")) == NULL ) {
 	error_e("could not fdopen");
 	goto exiterr;
     }
-#if ! (defined(HAVE_SETREGID) && defined(HAVE_SETREUID))
+#ifndef USE_SETE_ID
     if (fchown(file, asuid, asgid) != 0) {
-	error_e("Could not fchown %s to asuid and asgid", tmp);
+	error_e("Could not fchown %s to asuid and asgid", tmp_str);
 	goto exiterr;
     }
 #endif
@@ -492,7 +433,7 @@ edit_file(char *buf)
 
     do {
 
-	if ( stat(tmp, &st) == 0 )
+	if ( stat(tmp_str, &st) == 0 )
 	    mtime = st.st_mtime;
 	else {
 	    error_e("could not stat \"%s\"", buf);
@@ -502,29 +443,26 @@ edit_file(char *buf)
 	switch ( pid = fork() ) {
 	case 0:
 	    /* child */
-#if defined(HAVE_SETREGID) && defined(HAVE_SETREUID)
-	    if (uid != 0) {
-		if (setregid(asgid, asgid) < 0) {
-		    error_e("setregid(asgid, asgid)");
+	    if ( uid != ROOTUID ) {
+		if (setgid(asgid) < 0) {
+		    error_e("setgid(asgid)");
 		    goto exiterr;
 		}
-		if (setreuid(asuid, asuid) < 0) {
-		    error_e("setreuid(asuid, asuid)");
+		if (setuid(asuid) < 0) {
+		    error_e("setuid(asuid)");
 		    goto exiterr;
 		}
 	    }
-#else
-  	    if (setgid(asgid) < 0) {
-  		error_e("setgid(asgid)");
-  		goto exiterr;
-  	    }
-	    if (setuid(asuid) < 0) {
-		error_e("setuid(asuid)");
-		goto exiterr;
+	    else {
+		/* Some programs, like perl, require gid=egid : */
+		if ( setgid(getgid()) < 0 ) {
+		    error_e("setgid(getgid())");
+		    goto exiterr;
+		}
 	    }
-#endif
-	    execlp(editor, editor, tmp, NULL);
-	    error_e("Error while running \"%s\"", editor);
+	    snprintf(editorcmd, sizeof(editorcmd), "%s %s", cureditor, tmp_str);
+	    execlp(shell, shell, "-c", editorcmd, tmp_str, NULL);
+	    error_e("Error while running \"%s\"", cureditor);
 	    goto exiterr;
 
 	case -1:
@@ -537,16 +475,44 @@ edit_file(char *buf)
 	}
 	    
 	/* only reached by parent */
-	wait4(pid, &status, 0, NULL);
+	waitpid(pid, &status, 0);
 	if ( ! WIFEXITED(status) ) {
 	    fprintf(stderr, "Editor exited abnormally:"
 		    " fcrontab is unchanged.\n");
 	    goto exiterr;
 	}
 
+#ifndef USE_SETE_ID
+	/* we have chown the tmp file to user's name : user may have
+	 * linked the tmp file to a file owned by root. In that case, as
+	 * fcrontab is setuid root, user may read some informations he is not
+	 * allowed to read :
+	 * we *must* check that the tmp file is not a link. */
+
+	/* open the tmp file, chown it to root and chmod it to avoid race
+	 * conditions */
+	/* make sure that the tmp file is not a link */
+	{
+	    int fd = 0;
+	    if ( (fd = open(tmp_str, O_RDONLY)) <= 0 ||
+		 fstat(fd, &st) != 0 || ! S_ISREG(st.st_mode) ||
+		 S_ISLNK(st.st_mode) || st.st_uid != asuid || st.st_nlink > 1){
+		fprintf(stderr, "%s is not a valid regular file.\n", tmp_str);
+		close(fd);
+		goto exiterr;
+	    }
+	    if ( fchown(fd, ROOTUID, ROOTGID) != 0 || fchmod(fd, S_IRUSR|S_IWUSR) != 0 ){
+		fprintf(stderr, "Can't chown or chmod %s.\n", tmp_str);
+		close(fd);
+		goto exiterr;
+	    }
+	    close(fd);
+	}
+#endif
+	
 	/* check if file has been modified */
-	if ( stat(tmp, &st) != 0 ) {
-	    error_e("could not stat %s", tmp);
+	if ( stat(tmp_str, &st) != 0 ) {
+	    error_e("could not stat %s", tmp_str);
 	    goto exiterr;
 	}    
 
@@ -554,7 +520,7 @@ edit_file(char *buf)
 
 	    correction = 0;
 
-	    switch ( read_file(tmp, user) ) {
+	    switch ( read_file(tmp_str) ) {
 	    case ERR:
 		goto exiterr;
 	    case 2:
@@ -586,7 +552,8 @@ edit_file(char *buf)
 
     } while ( correction == 1);
 
-    write_file(tmp);
+    if ( write_file(tmp_str) != OK )
+	return_val = EXIT_ERR;
     
     /* free memory used to store the list */
     delete_file(user);
@@ -595,13 +562,15 @@ edit_file(char *buf)
     need_sig = 1;
     
   end:
-    if ( remove(tmp) != 0 )
-	error_e("could not remove %s", tmp);
-    xexit (EXIT_OK);
+    if ( remove(tmp_str) != 0 )
+	error_e("could not remove %s", tmp_str);
+    free(tmp_str);
+    xexit (return_val);
 
   exiterr:
-    if ( remove(tmp) != 0 )
-	error_e("could not remove %s", tmp);
+    if ( remove(tmp_str) != 0 )
+	error_e("could not remove %s", tmp_str);
+    free(tmp_str);
     xexit (EXIT_ERR);
 
 }
@@ -611,22 +580,24 @@ int
 install_stdin(void)
     /* install what we get through stdin */
 {
+    int tmp_fd = 0;
     FILE *tmp_file = NULL;
-    char tmp[FNAME_LEN];
-    register char c;
+    char *tmp_str = NULL;
+    int c;
     short return_val = EXIT_OK;
 	    	    
-    sprintf(tmp, "/tmp/fcrontab.%d", getpid());
-    if( (tmp_file = fopen(tmp, "w")) == NULL )
-	die_e("Could not open \"%s\"", tmp);
+    tmp_fd = temp_file(&tmp_str);
+    
+    if( (tmp_file = fdopen(tmp_fd, "w")) == NULL )
+	die_e("Could not fdopen file %s", tmp_str);
 
     while ( (c = getc(stdin)) != EOF )
 	putc(c, tmp_file);
 
     fclose(tmp_file);
+    close(tmp_fd);
 
-
-    if ( make_file(tmp) == ERR )
+    if ( make_file(tmp_str) == ERR )
 	goto exiterr;
     else
 	goto exit;
@@ -634,8 +605,9 @@ install_stdin(void)
   exiterr:
 	return_val = EXIT_ERR;    
   exit:
-    if ( remove(tmp) != 0 )
-	error_e("Could not remove %s", tmp);
+    if ( remove(tmp_str) != 0 )
+	error_e("Could not remove %s", tmp_str);
+    free(tmp_str);
     return return_val;
 
 }
@@ -650,13 +622,13 @@ reinstall(char *buf)
     if ( (i = open(buf, O_RDONLY)) < 0) {
 	if ( errno == ENOENT ) {
 	    fprintf(stderr, "Could not reinstall: user %s has no fcrontab\n",
-		    buf);
+		    user);
 	}
 	else
 	    fprintf(stderr, "Could not open \"%s\": %s\n", buf,
 		    strerror(errno));
 
-	exit (EXIT_ERR);
+	xexit(EXIT_ERR);
     }
 
     close(0); dup2(i, 0);
@@ -666,21 +638,121 @@ reinstall(char *buf)
 
 }
 
+
+#ifdef HAVE_LIBPAM
+int
+conv_pam(int num_msg, const struct pam_message **msgm, struct pam_response **response,
+	 void *appdata_ptr)
+    /* text based conversation for pam. */
+{
+    int count = 0;
+    struct pam_response *reply;
+
+    if (num_msg <= 0 )
+	return PAM_CONV_ERR;
+
+    reply = (struct pam_response *) calloc(num_msg, sizeof(struct pam_response));
+    if (reply == NULL) {
+	debug("no memory for responses");
+	return PAM_CONV_ERR;
+    }
+
+    for (count = 0; count < num_msg; ++count) {
+	char *string = NULL;
+
+	switch ( msgm[count]->msg_style ) {
+	case PAM_PROMPT_ECHO_OFF:
+	    string = read_string(CONV_ECHO_OFF,msgm[count]->msg);
+	    if (string == NULL) {
+		goto failed_conversation;
+	    }
+	    break;
+	case PAM_PROMPT_ECHO_ON:
+	    string = read_string(CONV_ECHO_ON,msgm[count]->msg);
+	    if (string == NULL) {
+		goto failed_conversation;
+	    }
+	    break;
+	case PAM_ERROR_MSG:
+	    if (fprintf(stderr,"%s\n",msgm[count]->msg) < 0) {
+		goto failed_conversation;
+	    }
+	    break;
+	case PAM_TEXT_INFO:
+	    if (fprintf(stdout,"%s\n",msgm[count]->msg) < 0) {
+		goto failed_conversation;
+	    }
+	    break;
+	default:
+	    fprintf(stderr, "erroneous conversation (%d)\n"
+		    ,msgm[count]->msg_style);
+	    goto failed_conversation;
+	}
+
+	if (string) {                         /* must add to reply array */
+	    /* add string to list of responses */
+
+	    reply[count].resp_retcode = 0;
+	    reply[count].resp = string;
+	    string = NULL;
+	}
+    }
+
+    /* New (0.59+) behavior is to always have a reply - this is
+       compatable with the X/Open (March 1997) spec. */
+    *response = reply;
+    reply = NULL;
+
+    return PAM_SUCCESS;
+
+failed_conversation:
+
+    if (reply) {
+	for (count=0; count<num_msg; ++count) {
+	    if (reply[count].resp == NULL) {
+		continue;
+	    }
+	    switch (msgm[count]->msg_style) {
+	    case PAM_PROMPT_ECHO_ON:
+	    case PAM_PROMPT_ECHO_OFF:
+		Overwrite(reply[count].resp);
+		free(reply[count].resp);
+		break;
+	    case PAM_ERROR_MSG:
+	    case PAM_TEXT_INFO:
+		/* should not actually be able to get here... */
+		free(reply[count].resp);
+	    }                                            
+	    reply[count].resp = NULL;
+	}
+	/* forget reply too */
+	free(reply);
+	reply = NULL;
+    }
+
+    return PAM_CONV_ERR;
+}
+#endif /* HAVE_LIBPAM */
+
+
 void
 parseopt(int argc, char *argv[])
   /* set options */
 {
 
-    char c;
+    int c;
     extern char *optarg;
     extern int optind, opterr, optopt;
     struct passwd *pass;
+#ifdef SYSFCRONTAB
+    char is_sysfcrontab = 0;
+#endif
 
     /* constants and variables defined by command line */
 
     while(1) {
-	c = getopt(argc, argv, "u:lrezdnhV");
-	if (c == -1) break;
+	c = getopt(argc, argv, "u:lrezdnhVc:");
+	if (c == EOF) break;
 	switch (c) {
 
 	case 'V':
@@ -690,11 +762,11 @@ parseopt(int argc, char *argv[])
 	    usage(); break;
 
 	case 'u':
-	    user = strdup2(optarg) ; 
-	    if (getuid() != 0) {
+	    if (uid != ROOTUID) {
 		fprintf(stderr, "must be privileged to use -u\n");
 		xexit(EXIT_ERR);
 	    }
+	    user = strdup2(optarg) ; 
 	    break;
 
 	case 'd':
@@ -707,9 +779,7 @@ parseopt(int argc, char *argv[])
 		xexit(EXIT_ERR);
 	    }
 	    list_opt = 1;
-	    rm_opt = 0;
-	    edit_opt = 0;
-	    reinstall_opt = 0;
+	    rm_opt = edit_opt = reinstall_opt = 0;
 	    break;
 
 	case 'r':
@@ -718,10 +788,8 @@ parseopt(int argc, char *argv[])
 			"may be used simultaneously.\n");
 		xexit(EXIT_ERR);
 	    }
-	    list_opt = 0;
 	    rm_opt = 1;
-	    edit_opt = 0;
-	    reinstall_opt = 0;
+	    list_opt = edit_opt = reinstall_opt = 0;
 	    break;
 
 	case 'e':
@@ -730,10 +798,8 @@ parseopt(int argc, char *argv[])
 			"may be used simultaneously.\n");
 		xexit(EXIT_ERR);
 	    }
-	    list_opt = 0;
-	    rm_opt = 0;
 	    edit_opt = 1;
-	    reinstall_opt = 0;
+	    list_opt = rm_opt = reinstall_opt = 0;
 	    break;
 
 	case 'z':
@@ -742,28 +808,69 @@ parseopt(int argc, char *argv[])
 			"may be used simultaneously.\n");
 		xexit(EXIT_ERR);
 	    }
-	    list_opt = 0;
-	    rm_opt = 0;
-	    edit_opt = 0;
-	    reinstall_opt = 1;
+	    reinstall_opt = ignore_prev = 1;
+	    list_opt = rm_opt = edit_opt = 0;
 	    break;
 
 	case 'n':
 	    ignore_prev = 1;
 	    break;
 	    
+	case 'c':
+	    if ( optarg[0] == '/' ) {
+		Set(fcronconf, optarg);
+	    }
+	    else {
+		char buf[PATH_LEN];
+		snprintf(buf, sizeof(buf), "%s/%s", orig_dir, optarg);
+		Set(fcronconf, buf);
+	    }
+	    break;
+
 	case ':':
-	    fprintf(stderr, "(setopt) Missing parameter");
+	    fprintf(stderr, "(setopt) Missing parameter.\n");
 	    usage();
 
 	case '?':
 	    usage();
 
 	default:
-	    fprintf(stderr, "(setopt) Warning: getopt returned %c", c);
+	    fprintf(stderr, "(setopt) Warning: getopt returned %c.\n", c);
 	}
     }
 
+    /* read fcron.conf and update global parameters */
+    read_conf();
+    
+    /* read the file name and/or user and check validity of the arguments */
+    if (argc - optind > 2)
+	usage();
+    else if (argc - optind == 2 ) {
+	if ( list_opt + rm_opt + edit_opt + reinstall_opt == 0 )
+	    file_opt = optind++;
+	else
+	    usage();
+
+	if (uid != ROOTUID) {
+	    fprintf(stderr, "must be privileged to use -u\n");
+	    xexit(EXIT_ERR);
+	}
+	Set(user, argv[optind]); 
+    }
+    else if (argc - optind == 1) {
+	if ( list_opt + rm_opt + edit_opt + reinstall_opt == 0 )
+	    file_opt = optind;
+	else {
+	    if (uid != ROOTUID) {
+		fprintf(stderr, "must be privileged to use [user|-u user]\n");
+		xexit(EXIT_ERR);
+	    }
+	    Set(user, argv[optind]); 	    
+	}
+    }
+    else if (list_opt + rm_opt + edit_opt + reinstall_opt != 1)
+	usage();
+
     if ( user == NULL ) {
 	/* get user's name using getpwuid() */
 	if ( ! (pass = getpwuid(uid)) )
@@ -775,19 +882,38 @@ parseopt(int argc, char *argv[])
 	asgid = pass->pw_gid;
     }
     else {
-	if ( ! (pass = getpwnam(user)) )
-	    die("user \"%s\" is not in passwd file. Aborting.", user);
-	asuid = pass->pw_uid;
-	asgid = pass->pw_gid;
+#ifdef SYSFCRONTAB
+  	if ( strcmp(user, SYSFCRONTAB) == 0 ) {
+	    is_sysfcrontab = 1;
+	    asuid = ROOTUID;
+	    asgid = ROOTGID;
+	}
+	else
+#endif /* def SYSFCRONTAB */ 
+	    if ( (pass = getpwnam(user)) ) {
+		asuid = pass->pw_uid;
+		asgid = pass->pw_gid;
+	    }
+	    else
+		die("user \"%s\" is not in passwd file. Aborting.", user);
     }
 
-    if ( ! is_allowed(user) ) {
+    if ( 
+#ifdef SYSFCRONTAB
+	! is_sysfcrontab &&
+#endif
+	! is_allowed(user) ) {
 	die("User \"%s\" is not allowed to use %s. Aborting.",
 	    user, prog_name);	    
     }
 
-    if (optind < argc)
-	file_opt = optind;
+#ifdef SYSFCRONTAB
+    if ( is_sysfcrontab )
+	runas = ROOTNAME;
+    else
+#endif
+	runas = user;
+
 }
 
 
@@ -795,64 +921,102 @@ int
 main(int argc, char **argv)
 {
 
+#ifdef HAVE_LIBPAM
+    int    retcode = 0;
+    const char * const * env;
+#endif
+#ifdef USE_SETE_ID
+    struct passwd *pass;
+#endif
+
     memset(buf, 0, sizeof(buf));
     memset(file, 0, sizeof(file));
 
     if (strrchr(argv[0],'/')==NULL) prog_name = argv[0];
     else prog_name = strrchr(argv[0],'/')+1;
-
+    
     uid = getuid();
 
     /* get current dir */
-    if ( (orig_dir = getcwd(NULL, 0)) == NULL )
-	error_e("getcwd");
+    if ( getcwd(orig_dir, sizeof(orig_dir)) == NULL )
+	die_e("getcwd");
 
     /* interpret command line options */
     parseopt(argc, argv);
 
-#if defined(HAVE_SETREGID) && defined(HAVE_SETREUID)
-    {
-	struct passwd *pass;
-	if ( ! (pass = getpwnam(USERNAME)) )
-	    die("user \"%s\" is not in passwd file. Aborting.", USERNAME);
-	fcrontab_uid = pass->pw_uid;
-	fcrontab_gid = pass->pw_gid;
+#ifdef USE_SETE_ID
+    if ( ! (pass = getpwnam(USERNAME)) )
+	die("user \"%s\" is not in passwd file. Aborting.", USERNAME);
+    fcrontab_uid = pass->pw_uid;
+    fcrontab_gid = pass->pw_gid;
 
-	if (uid != fcrontab_uid)
-	    if (seteuid(fcrontab_uid) != 0) 
-		die_e("Could not change uid to fcrontab_uid[%d]",fcrontab_uid);
-	/* change directory */
-	if (chdir(cdir) != 0) {
-	    error_e("Could not chdir to " FCRONTABS );
-	    xexit (EXIT_ERR);
-	}
-	/* get user's permissions */
-	if (seteuid(uid) != 0) 
-	    die_e("Could not change uid to uid[%d]", uid); 
-	if (setegid(fcrontab_gid) != 0) 
-	    die_e("Could not change gid to " GROUPNAME "[%d]", fcrontab_gid); 
+#ifdef HAVE_LIBPAM
+    /* Open PAM session for the user and obtain any security
+       credentials we might need */
+
+    debug("username: %s", user);
+    retcode = pam_start("fcrontab", user, &apamconv, &pamh);
+    if (retcode != PAM_SUCCESS) die_pame(pamh, retcode, "Could not start PAM");
+    retcode = pam_authenticate(pamh, 0);    /* is user really user? */
+    if (retcode != PAM_SUCCESS)
+	die_pame(pamh, retcode, "Could not authenticate user using PAM (%d)", retcode);
+    retcode = pam_acct_mgmt(pamh, 0); /* permitted access? */
+    if (retcode != PAM_SUCCESS)
+	die_pame(pamh, retcode, "Could not init PAM account management (%d)", retcode);
+    retcode = pam_setcred(pamh, PAM_ESTABLISH_CRED);
+    if (retcode != PAM_SUCCESS) die_pame(pamh, retcode, "Could not set PAM credentials");
+    retcode = pam_open_session(pamh, 0);
+    if (retcode != PAM_SUCCESS) die_pame(pamh, retcode, "Could not open PAM session");
+
+    env = (const char * const *) pam_getenvlist(pamh);
+    while (env && *env) {
+	if (putenv((char*) *env)) die_e("Could not copy PAM environment");
+	env++;
     }
-#else
-    if (setuid(0) != 0 ) 
-	die_e("Could not change uid to 0"); 
-    if (setgid(0) != 0)
-    	die_e("Could not change gid to 0");
+
+    /* Close the log here, because PAM calls openlog(3) and
+       our log messages could go to the wrong facility */
+    xcloselog();
+#endif /* USE_PAM */
+
+    if (uid != fcrontab_uid)
+	if (seteuid(fcrontab_uid) != 0) 
+	    die_e("Couldn't change euid to fcrontab_uid[%d]",fcrontab_uid);
     /* change directory */
-    if (chdir(cdir) != 0) {
-	error_e("Could not chdir to " FCRONTABS );
+    if (chdir(fcrontabs) != 0) {
+	error_e("Could not chdir to %s", fcrontabs);
 	xexit (EXIT_ERR);
     }
-#endif
-    
+    /* get user's permissions */
+    if (seteuid(uid) != 0) 
+	die_e("Could not change euid to %d", uid); 
+    if (setegid(fcrontab_gid) != 0) 
+	die_e("Could not change egid to " GROUPNAME "[%d]", fcrontab_gid); 
 
+#else /* USE_SETE_ID */
+
+    if (setuid(ROOTUID) != 0 ) 
+	die_e("Could not change uid to ROOTUID"); 
+    if (setgid(ROOTGID) != 0)
+    	die_e("Could not change gid to ROOTGID");
+    /* change directory */
+    if (chdir(fcrontabs) != 0) {
+	error_e("Could not chdir to %s", fcrontabs);
+	xexit (EXIT_ERR);
+    }
+#endif /* USE_SETE_ID */
+    
     /* this program is seteuid : we set default permission mode
-     * to  600 for security reasons */
-    umask(026);
+     * to 640 for a normal user, 600 for root, for security reasons */
+    if ( asuid == ROOTUID )
+	umask(066);  /* octal : '0' + number in octal notation */
+    else
+	umask(026);
 
     snprintf(buf, sizeof(buf), "%s.orig", user);
 
     /* determine what action should be taken */
-    if ( file_opt && ! list_opt && ! rm_opt && ! edit_opt && ! reinstall_opt) {
+    if ( file_opt ) {
 
 	if ( strcmp(argv[file_opt], "-") == 0 )
 
@@ -862,62 +1026,47 @@ main(int argc, char **argv)
 
 	    if ( *argv[file_opt] != '/' )
 		/* this is just the file name, not the path : complete it */
-		snprintf(file,sizeof(file),"%s/%s",orig_dir,argv[file_opt]);
-	    else
+		snprintf(file, sizeof(file), "%s/%s", orig_dir, argv[file_opt]);
+	    else {
 		strncpy(file, argv[file_opt], sizeof(file) - 1);
+		file[sizeof(file)-1] = '\0';
+	    }
 
 	    if (make_file(file) == OK)
-		xexit ( EXIT_OK );
+		xexit(EXIT_OK);
 	    else
-		xexit ( EXIT_ERR );
+		xexit(EXIT_ERR);
 
 	}
 
     } 
-    else if(list_opt + rm_opt + edit_opt + reinstall_opt != 1 || file_opt != 0)
-	usage();
-
 
     /* remove user's entries */
     if ( rm_opt == 1 ) {
-
 	if ( remove_fcrontab(1) == ENOENT )
 	    fprintf(stderr, "no fcrontab for %s\n", user);
-
 	xexit (EXIT_OK);
     }
 
-
     /* list user's entries */
     if ( list_opt == 1 ) {
-
 	list_file(buf);
-
 	xexit(EXIT_OK);
-
     }
 
 
     /* edit user's entries */
     if ( edit_opt == 1 ) {
-
 	edit_file(buf);
-
 	xexit(EXIT_OK);
-
     }
 
     /* reinstall user's entries */
     if ( reinstall_opt == 1 ) {
-
 	reinstall(buf);
-
 	xexit(EXIT_OK);
-
     }
 
-
-
-    /* never reach */
+    /* never reached */
     return EXIT_OK;
 }
