@@ -2,7 +2,7 @@
 /*
  * FCRON - periodic command scheduler 
  *
- *  Copyright 2000-2001 Thibault Godouet <fcron@free.fr>
+ *  Copyright 2000-2004 Thibault Godouet <fcron@free.fr>
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -22,33 +22,32 @@
  *  `LICENSE' that comes with the fcron source distribution.
  */
 
- /* $Id: fileconf.c,v 1.42 2001/03/10 13:10:23 thib Exp $ */
+ /* $Id: fileconf.c,v 1.72 2003/12/29 20:00:40 thib Exp thib $ */
 
 #include "fcrontab.h"
 
+#include "fileconf.h"
+
 char *get_string(char *ptr);
 int get_line(char *str, size_t size, FILE *file);
-char *get_time(char *ptr, time_t *time);
+char *get_time(char *ptr, time_t *time, int zero_allowed);
 char *get_num(char *ptr, int *num, int max, short int decimal,
 	      const char **names);
 char *get_nice(char *ptr, int *nice);
 char *get_bool(char *ptr, int *i);
 char *read_field(char *ptr, bitstr_t *ary, int max, const char **names);
-void read_freq(char *ptr, CF *cf);
-void read_arys(char *ptr, CF *cf);
-void read_period(char *ptr, CF *cf);
-void read_env(char *ptr, CF *cf);
-char *read_opt(char *ptr, CL *cl);
+void read_freq(char *ptr, cf_t *cf);
+void read_arys(char *ptr, cf_t *cf);
+void read_period(char *ptr, cf_t *cf);
+void read_env(char *ptr, cf_t *cf);
+char *read_opt(char *ptr, cl_t *cl);
+char *check_username(char *ptr, cf_t *cf, cl_t *cl);
 
 
 char need_correction;
-CL default_line;    /* default options for a line */
+cl_t default_line;    /* default options for a line */
 char *file_name;
 int line;
-extern char *user;
-extern uid_t uid;
-extern uid_t asuid;
-extern uid_t fcrontab_uid;
 
 /* warning : all names must have the same length */
 const char *dows_ary[] = {
@@ -102,11 +101,12 @@ get_line(char *str, size_t size, FILE *file)
      * and continue over an "\" followed by an "\n" char */
 {
     size_t size_max = size - 1 ;
-    register int i=0;
+    int i=0;
+    int c;
 
     while (i < size_max ) {
 
-	switch ( *(str + i) = getc(file) ) {
+	switch ( c = getc(file) ) {
 
 	case '\n':
 	    /* check if the \n char is preceded by a "\" char : 
@@ -118,18 +118,19 @@ get_line(char *str, size_t size, FILE *file)
 		continue;
 	    }
 	    else {
-		*(str + i) = '\0';
+		*(str + i) = (char) '\0';
 		return OK;
 	    }
 	    break;
 		
 	case EOF:
-	    *(str + i) = '\0';
+	    *(str + i) = (char) '\0';
 	    /* we couldn't return EOF ( equal to ERR by default ) 
 	     * nor ERR, which is used for another error */
 	    return 999;
 
 	default:
+	    *(str + i) = (char) c;
 	    i++;
 
 	}
@@ -137,7 +138,7 @@ get_line(char *str, size_t size, FILE *file)
     }
 
     /* line is too long : goto next line and return ERR */
-    while ((*str = getc(file)) != '\n' && *str != EOF )
+    while ( ( (char)(c = getc(file)) != EOF ) && ( (char)c != '\n') )
 	;
     line++;
     need_correction = 1;
@@ -146,10 +147,10 @@ get_line(char *str, size_t size, FILE *file)
 }
 
 int
-read_file(char *filename, char *user)
-    /* read file "name" and append CF list */
+read_file(char *filename)
+    /* read file "name" and append cf_t list */
 {
-    CF *cf = NULL;
+    cf_t *cf = NULL;
     FILE *file = NULL;
     char buf[LINE_LEN];
     int max_lines;
@@ -159,7 +160,7 @@ read_file(char *filename, char *user)
     int ret;
     
     bzero(buf, sizeof(buf));
-    bzero(&default_line, sizeof(CL));
+    bzero(&default_line, sizeof(cl_t));
     need_correction = 0;
     line = 1;
     file_name = filename;
@@ -167,7 +168,6 @@ read_file(char *filename, char *user)
     /* open file */
 
     /* check if user is allowed to read file */
-    /* create a temp file with user's permissions */
     if ( access(file_name, R_OK) != 0 )
 	die_e("User %s can't read file \"%s\"", user, file_name);
     else if ( (file = fopen(file_name, "r")) == NULL ) {
@@ -176,16 +176,17 @@ read_file(char *filename, char *user)
 	return ERR;
     }
 
-    Alloc(cf, CF);
+    Alloc(cf, cf_t);
     cf->cf_user = strdup2(user);
     default_line.cl_file = cf;
-    default_line.cl_runas = strdup2(user);
-    default_line.cl_mailto = strdup2(user);
+    default_line.cl_runas = strdup2(runas);
+    default_line.cl_mailto = strdup2(runas);
+    set_default_opt(default_line.cl_option);
 
     if ( debug_opt )
 	fprintf(stderr, "FILE %s\n", file_name);
 
-    if (strcmp(user, "root") == 0)
+    if (strcmp(runas, "root") == 0)
 	max_entries = 65535;
 
     /* max_lines acts here as a security counter to avoid endless loop. */
@@ -193,16 +194,12 @@ read_file(char *filename, char *user)
 
     while ( entries <= max_entries && line <= max_lines ) {
 
-	if ( (ret = get_line(buf, sizeof(buf), file)) == OK)
-	    ;
-	else if ( ret == ERR ) {
-	    fprintf(stderr, "%s:%d: Line is too long (more than %d):"
-		    " skipping line.\n", file_name, line, sizeof(buf));
+	if ( (ret = get_line(buf, sizeof(buf), file)) == ERR ) {
+	    fprintf(stderr, "%s:%d: Line is too long (more than %d): skipping line.\n",
+		    file_name, line, sizeof(buf));
 	    continue;
-	} else
-	    /* EOF : no more lines */
-	    break;
-	
+	}
+
 	ptr = buf;
 	Skip_blanks(ptr);
 
@@ -236,7 +233,7 @@ read_file(char *filename, char *user)
 	    }
 	    break;
 	default:
-	    if ( isdigit(*ptr) || *ptr == '*' ) {
+	    if ( isdigit( (int) *ptr) || *ptr == '*' ) {
 		read_arys(ptr, cf);
 		entries++;
 	    } else
@@ -245,6 +242,10 @@ read_file(char *filename, char *user)
 
 	line++;	
 
+	if ( ret != OK )
+	    /* in this case, ret == EOF : no more lines */
+	    break;
+	
     }
 
     if (entries == max_entries) {
@@ -272,7 +273,7 @@ read_file(char *filename, char *user)
 }
 
 void
-read_env(char *ptr, CF *cf)
+read_env(char *ptr, cf_t *cf)
     /* append env variable list.
      * (remove blanks) */
 {
@@ -284,8 +285,8 @@ read_env(char *ptr, CF *cf)
     bzero(name, sizeof(name));
 
     /* copy env variable's name */
-    while ( (isalnum(*ptr) || *ptr == '_') && *ptr != '=' && !isspace(*ptr) 
-	    && j < sizeof(name) ) {
+    while ( (isalnum( (int) *ptr) || *ptr == '_') && *ptr != '=' 
+	    && ! isspace( (int) *ptr) && j < sizeof(name) ) {
 	name[j++] = *ptr;
 	ptr++;
     }
@@ -295,14 +296,14 @@ read_env(char *ptr, CF *cf)
 	goto error;
 
     /* skip '=' and spaces around */
-    while ( isspace(*ptr) )
+    while ( isspace( (int) *ptr) )
 	ptr++;
 
     /* if j == 0 name is a zero length string */
     if ( *ptr++ != '=' || j == 0 )
 	goto error;
 
-    while ( isspace(*ptr) )
+    while ( isspace( (int) *ptr) )
 	ptr++;
 
     /* get value */
@@ -325,20 +326,13 @@ read_env(char *ptr, CF *cf)
 
     /* the MAILTO assignment is, in fact, an fcron option :
      *  we don't store it in the same way. */
+    /* please note that we check if the mailto is valid in conf.c */
     if ( strcmp(name, "MAILTO") == 0 ) {
 	if ( strcmp(val, "\0") == 0 )
 	    clear_mail(default_line.cl_option);
 	else {
-	    struct passwd *pass = NULL;
-	    if ( (pass = getpwnam(val)) == 0 ) {
-		fprintf(stderr, "%s:%d:MAILTO: \"%s\" is not in passwd :"
-			" ignored\n", file_name, line, val);	
-		need_correction = 1;
-	    } else {
-		free(default_line.cl_mailto);
-		default_line.cl_mailto = strdup2(val);
-		set_mail(default_line.cl_option);
-	    }
+	    Set(default_line.cl_mailto, val);
+	    set_mail(default_line.cl_option);
 	}
 	    
     }
@@ -378,7 +372,7 @@ get_nice(char *ptr, int *nice)
 	return NULL;
 
     if ( negative == 1 ) {
-	if (getuid() != 0) {
+	if (getuid() != ROOTUID) {
 	    fprintf(stderr, "must be privileged to use a negative argument "
 		    "with nice: set to 0\n");
 	    need_correction = 1;
@@ -435,7 +429,7 @@ get_bool(char *ptr, int *i)
 
 
 char *
-read_opt(char *ptr, CL *cl)
+read_opt(char *ptr, cl_t *cl)
     /* read one or several options and fill in the field "option" */
 {
     char opt_name[20];
@@ -457,7 +451,7 @@ read_opt(char *ptr, CL *cl)
 	i = 0;
 	bzero(opt_name, sizeof(opt_name));
 
-	while ( isalnum(*ptr)  && i < sizeof(opt_name))
+	while ( isalnum( (int) *ptr)  && i < sizeof(opt_name))
 	    opt_name[i++] = *ptr++;
     
 	i = 1;
@@ -468,7 +462,30 @@ read_opt(char *ptr, CL *cl)
 	    ptr++;
 	}
 
-	if ( strcmp(opt_name, "s") == 0 || strcmp(opt_name, "serial") == 0 ) {
+	/* global options for a file */
+	if ( strcmp(opt_name, "tzdiff") == 0 ) {
+	    char negative = 0;
+
+	    if ( ! in_brackets )
+		Handle_err;
+	    if ( *ptr == '-' ) {
+		negative = 1;
+		ptr++;
+	    }
+	    if ( (ptr = get_num(ptr, &i, 24, 0, NULL)) == NULL )
+		Handle_err;
+	    if ( negative )
+		cl->cl_file->cf_tzdiff = (- i);
+	    else
+		cl->cl_file->cf_tzdiff = i;
+
+ 	    if (debug_opt)
+		fprintf(stderr, "  Opt : \"%s\" (-)%d\n", opt_name, i);
+	}
+	
+
+	/* options related to a line (or a set of lines) */
+	else if(strcmp(opt_name, "s") == 0 || strcmp(opt_name, "serial") == 0){
 	    if ( in_brackets && (ptr = get_bool(ptr, &i)) == NULL )
 		Handle_err;
 	    if (i == 0 )
@@ -527,129 +544,39 @@ read_opt(char *ptr, CL *cl)
 	    if ( in_brackets && (ptr = get_bool(ptr, &i)) == NULL )
 		Handle_err;
 	    if ( i == 1 ) {
-		bzero(cl, sizeof(CL));
-		free(cl->cl_runas);
-		cl->cl_runas = strdup2(user);
-		free(cl->cl_mailto);
-		cl->cl_mailto = strdup2(user);
+		bzero(cl, sizeof(cl_t));
+		Set(cl->cl_runas, runas);
+		Set(cl->cl_mailto, runas);
+		set_default_opt(cl->cl_option);
 	    }
 	    if (debug_opt)
 		fprintf(stderr, "  Opt : \"%s\"\n", opt_name);
 	}
 
-	else if(strcmp(opt_name, "f") == 0 || strcmp(opt_name, "first") == 0){
-	    if( ! in_brackets || (ptr=get_time(ptr, &(cl->cl_nextexe)))==NULL)
+	else if(strcmp(opt_name, "f") == 0 || strcmp(opt_name, "first") == 0) {
+	    if( ! in_brackets || (ptr=get_time(ptr, &(cl->cl_first), 1)) == NULL )
 		Handle_err;
  	    if (debug_opt)
-		fprintf(stderr, "  Opt : \"%s\" %ld\n",opt_name,cl->cl_nextexe);
+		fprintf(stderr, "  Opt : \"%s\" %ld\n", opt_name,
+			(long int)cl->cl_first);
 	}
 
 	else if(strcmp(opt_name, "r")==0 || strcmp(opt_name, "runfreq")==0) {
-	    if( !in_brackets || (ptr=get_num(ptr,&i,USHRT_MAX,0,NULL)) == NULL)
+	    if ( cl->cl_runfreq == 1 ) {
+		fprintf(stderr, "cannot change runfreq value in a %%-line");
 		Handle_err;
-	    if (i <= 1)
+	    }
+	    if ( !in_brackets || (ptr=get_num(ptr,&i,USHRT_MAX,0,NULL)) == NULL)
+		Handle_err;
+	    if (i <= 1) {
 		fprintf(stderr, "runfreq must be 2 or more.\n");
-	    if (i <= 1)
-		Handle_err;	
+		Handle_err;
+	    }
 	    cl->cl_runfreq = i;
  	    if (debug_opt)
 		fprintf(stderr, "  Opt : \"%s\" %d\n", opt_name, i);
 	}
 
-	/* options to run once per interval :
-	 * ignore every fields below the limit */
-	else if (strcmp(opt_name, "mins") == 0) {
-	    /* nothing to do */
- 	    if (debug_opt)
-		fprintf(stderr, "  Opt : \"%s\"\n", opt_name);
-	}
-	else if (strcmp(opt_name, "hours") == 0) {
-	    set_freq_mins(cl->cl_option);
- 	    if (debug_opt)
-		fprintf(stderr, "  Opt : \"%s\"\n", opt_name);
-	}
-	else if (strcmp(opt_name, "days") == 0) {
-	    set_freq_mins(cl->cl_option);
-	    set_freq_hrs(cl->cl_option);
- 	    if (debug_opt)
-		fprintf(stderr, "  Opt : \"%s\"\n", opt_name);
-	}
-	else if (strcmp(opt_name, "mons") == 0) {
-	    set_freq_mins(cl->cl_option);
-	    set_freq_hrs(cl->cl_option);
-	    set_freq_days(cl->cl_option);
- 	    if (debug_opt)
-		fprintf(stderr, "  Opt : \"%s\"\n", opt_name);
-	}
-	else if (strcmp(opt_name, "dow") == 0) {
-	    set_freq_mins(cl->cl_option);
-	    set_freq_hrs(cl->cl_option);
-	    set_freq_days(cl->cl_option);
-	    set_freq_mons(cl->cl_option);
- 	    if (debug_opt)
-		fprintf(stderr, "  Opt : \"%s\"\n", opt_name);
-	}
-
-	/* run once an element of the selected field 
-	 * (once an hour, once a day, etc) */
-	else if (strcmp(opt_name, "hourly") == 0) {
-	    set_freq_hrs(cl->cl_option);
-	    set_freq_periodically(cl->cl_option);
- 	    if (debug_opt)
-		fprintf(stderr, "  Opt : \"%s\"\n", opt_name);
-	}
-	else if (strcmp(opt_name, "daily") == 0) {
-	    set_freq_days(cl->cl_option);
-	    set_freq_periodically(cl->cl_option);
- 	    if (debug_opt)
-		fprintf(stderr, "  Opt : \"%s\"\n", opt_name);
-	}
-	else if (strcmp(opt_name, "monthly") == 0) {
-	    set_freq_mons(cl->cl_option);
-	    set_freq_periodically(cl->cl_option);
- 	    if (debug_opt)
-		fprintf(stderr, "  Opt : \"%s\"\n", opt_name);
-	}
-	else if (strcmp(opt_name, "weekly") == 0) {
-	    set_freq_dow(cl->cl_option);
-	    set_freq_periodically(cl->cl_option);
- 	    if (debug_opt)
-		fprintf(stderr, "  Opt : \"%s\"\n", opt_name);
-	}
-
-	/* run once an element of the selected field 
-	 * from middle to middle of that field 
-	 * (ie once from 12h to 12h the following day) */
-	else if (strcmp(opt_name, "midhourly") == 0) {
-	    set_freq_hrs(cl->cl_option);
-	    set_freq_periodically(cl->cl_option);
-	    set_freq_mid(cl->cl_option);
- 	    if (debug_opt)
-		fprintf(stderr, "  Opt : \"%s\"\n", opt_name);
-	}
-	else if (strcmp(opt_name, "middaily") == 0 
-		 || strcmp(opt_name, "nightly") == 0) {
-	    set_freq_days(cl->cl_option);
-	    set_freq_periodically(cl->cl_option);
-	    set_freq_mid(cl->cl_option);
- 	    if (debug_opt)
-		fprintf(stderr, "  Opt : \"%s\"\n", opt_name);
-	}
-	else if (strcmp(opt_name, "midmonthly") == 0) {
-	    set_freq_mons(cl->cl_option);
-	    set_freq_periodically(cl->cl_option);
-	    set_freq_mid(cl->cl_option);
- 	    if (debug_opt)
-		fprintf(stderr, "  Opt : \"%s\"\n", opt_name);
-	}
-	else if (strcmp(opt_name, "midweekly") == 0) {
-	    set_freq_dow(cl->cl_option);
-	    set_freq_periodically(cl->cl_option);
-	    set_freq_mid(cl->cl_option);
- 	    if (debug_opt)
-		fprintf(stderr, "  Opt : \"%s\"\n", opt_name);
-	}
-
 	else if ( strcmp(opt_name, "strict") == 0 ) {
 	    if ( in_brackets && (ptr = get_bool(ptr, &i)) == NULL )
 		Handle_err;
@@ -690,20 +617,26 @@ read_opt(char *ptr, CL *cl)
 	    if(!in_brackets || (ptr=get_num(ptr,&i,UCHAR_MAX,1,NULL)) == NULL )
 		Handle_err;
 	    cl->cl_lavg[2] = i;
-	    set_lavg(cl->cl_option);
+ 	    if (debug_opt)
+		fprintf(stderr, "  Opt : 'lavg15' %d\n", i);
+	    if ( cl->cl_lavg[0] || cl->cl_lavg[1] || cl->cl_lavg[2] )
+		set_lavg(cl->cl_option);
+	    else
+		clear_lavg(cl->cl_option);
 #ifdef NOLOADAVG
 	    warn("As fcron has been compiled with no procfs support,\n"
 		 "you will not be able to use the lavg* options");
 #endif /* NOLOADAVG */ 
- 	    if (debug_opt)
-		fprintf(stderr, "  Opt : 'lavg15' %d\n", i);
 	}
 
 	else if( strcmp(opt_name, "lavg1") == 0 ) {
 	    if(!in_brackets ||(ptr=get_num(ptr, &i, UCHAR_MAX, 1, NULL))==NULL)
 		Handle_err;
 	    cl->cl_lavg[0] = i;
-	    set_lavg(cl->cl_option);
+	    if ( cl->cl_lavg[0] || cl->cl_lavg[1] || cl->cl_lavg[2] )
+		set_lavg(cl->cl_option);
+	    else
+		clear_lavg(cl->cl_option);
 #if NOLOADAVG
 	    warn("As fcron has been compiled with no procfs support,\n"
 		 "you will not be able to use the lavg* options");
@@ -716,7 +649,10 @@ read_opt(char *ptr, CL *cl)
 	    if(!in_brackets ||(ptr=get_num(ptr, &i, UCHAR_MAX, 1, NULL))==NULL)
 		Handle_err;
 	    cl->cl_lavg[1] = i;
-	    set_lavg(cl->cl_option);
+	    if ( cl->cl_lavg[0] || cl->cl_lavg[1] || cl->cl_lavg[2] )
+		set_lavg(cl->cl_option);
+	    else
+		clear_lavg(cl->cl_option);
 #ifdef NOLOADAVG
 	    warn("As fcron has been compiled with no procfs support,\n"
 		 "you will not be able to use the lavg* options");
@@ -729,7 +665,10 @@ read_opt(char *ptr, CL *cl)
 	    if(!in_brackets ||(ptr=get_num(ptr, &i, UCHAR_MAX, 1, NULL))==NULL)
 		Handle_err;
 	    cl->cl_lavg[2] = i;
-	    set_lavg(cl->cl_option);
+	    if ( cl->cl_lavg[0] || cl->cl_lavg[1] || cl->cl_lavg[2] )
+		set_lavg(cl->cl_option);
+	    else
+		clear_lavg(cl->cl_option);
 #ifdef NOLOADAVG
 	    warn("As fcron has been compiled with no procfs support,\n"
 		 "you will not be able to use the lavg* options");
@@ -760,11 +699,12 @@ read_opt(char *ptr, CL *cl)
 		fprintf(stderr, "  Opt : \"%s\" %d\n", opt_name, i);
 	}
 
-	else if(strcmp(opt_name, "u") == 0 || strcmp(opt_name, "until") == 0){
-	    if( ! in_brackets || (ptr=get_time(ptr, &(cl->cl_until)))==NULL)
+	else if(strcmp(opt_name, "u") == 0 || strcmp(opt_name, "until") == 0) {
+	    if( ! in_brackets || (ptr=get_time(ptr, &(cl->cl_until), 0)) == NULL )
 		Handle_err;
  	    if (debug_opt)
-		fprintf(stderr, "  Opt : \"%s\" %ld\n",opt_name,cl->cl_until);
+		fprintf(stderr, "  Opt : \"%s\" %ld\n", opt_name,
+			(long int)cl->cl_until);
 	}
 
 	else if(strcmp(opt_name, "m")==0 || strcmp(opt_name, "mail")==0){
@@ -791,27 +731,20 @@ read_opt(char *ptr, CL *cl)
 
 	else if( strcmp(opt_name, "mailto") == 0) {
 	    char buf[50];
-	    struct passwd *pass = NULL;
 	    bzero(buf, sizeof(buf));
 
 	    if( ! in_brackets )
 		Handle_err;
 
+	    /* please note that we check if the mailto is valid in conf.c */
 	    i = 0;
-	    while ( isalnum(*ptr) )
+	    while ( *ptr != ')' && i + 1 < sizeof(buf) )
 		buf[i++] = *ptr++;
 	    if ( strcmp(buf, "\0") == 0 )
 		clear_mail(cl->cl_option);
 	    else {
-		if ( (pass = getpwnam(buf)) == NULL ) {
-		    fprintf(stderr, "%s:%d:mailto: \"%s\" is not in passwd :"
-			    " ignored\n", file_name, line, buf);	
-		    need_correction = 1;
-		} else {
-		    free(cl->cl_mailto);
-		    cl->cl_mailto = strdup2(buf);
+		    Set(cl->cl_mailto, buf);
 		    set_mail(cl->cl_option);
-		}
 	    }
  	    if (debug_opt)
 		fprintf(stderr, "  Opt : \"%s\" \"%s\"\n", opt_name, buf);
@@ -839,7 +772,7 @@ read_opt(char *ptr, CL *cl)
 		fprintf(stderr, "  Opt : \"%s\" %d\n", opt_name, i);
 	}
 
-	else if( strcmp(opt_name, "nolog") == 0 ) {
+	else if ( strcmp(opt_name, "nolog") == 0 ) {
 	    if ( in_brackets && (ptr = get_bool(ptr, &i)) == NULL )
 		Handle_err;
 	    if ( i == 0 )
@@ -850,16 +783,38 @@ read_opt(char *ptr, CL *cl)
 		fprintf(stderr, "  Opt : \"%s\" %d\n", opt_name, i);
 	}
 
+ 	else if ( strcmp(opt_name, "volatile") == 0 ) {
+ 	    if ( in_brackets && (ptr = get_bool(ptr, &i)) == NULL )
+ 		Handle_err;
+ 	    if ( i == 0 )
+ 		clear_volatile(cl->cl_option);
+ 	    else
+ 		set_volatile(cl->cl_option);	
+  	    if (debug_opt)
+ 		fprintf(stderr, "  Opt : \"%s\" %d\n", opt_name, i);
+ 	}
+ 
+ 	else if ( strcmp(opt_name, "stdout") == 0 ) {
+ 	    if ( in_brackets && (ptr = get_bool(ptr, &i)) == NULL )
+ 		Handle_err;
+ 	    if ( i == 0 )
+ 		clear_stdout(cl->cl_option);
+ 	    else
+ 		set_stdout(cl->cl_option);	
+  	    if (debug_opt)
+ 		fprintf(stderr, "  Opt : \"%s\" %d\n", opt_name, i);
+ 	}
+ 
 	else if(strcmp(opt_name, "n") == 0 || strcmp(opt_name, "nice") == 0) {
 	    if( ! in_brackets || (ptr = get_nice(ptr, &i)) == NULL )
 		Handle_err;
 	    cl->cl_nice = (char)i;
  	    if (debug_opt)
-		fprintf(stderr, "  Opt : \"%s\" %d\n", opt_name, i);
+		fprintf(stderr, "  Opt : \"%s\" (-)%d\n", opt_name, i);
 	}
 
 	else if(strcmp(opt_name, "runas") == 0) {
-	    if (getuid() != 0) {
+	    if (getuid() != ROOTUID) {
 		fprintf(stderr, "must be privileged to use option runas: "
 			"skipping option\n");
 		need_correction = 1;
@@ -878,7 +833,7 @@ read_opt(char *ptr, CL *cl)
 
 		bzero(name, sizeof(name));
 
-		while ( *ptr != ')' && i < sizeof(name))
+		while ( *ptr != ')' && i + 1 < sizeof(name))
 		    name[i++] = *ptr++;
     
 		if ((pas = getpwnam(name)) == NULL) {
@@ -887,8 +842,7 @@ read_opt(char *ptr, CL *cl)
 		    need_correction = 1;
 		} 
 		else {
-		    free(cl->cl_runas);
-		    cl->cl_runas = strdup2(name);
+		    Set(cl->cl_runas, name);
 		    if (debug_opt)
 			fprintf(stderr, "  Opt : \"%s\" %s\n", opt_name, name);
 		}
@@ -896,6 +850,116 @@ read_opt(char *ptr, CL *cl)
 	    }
 	}
 
+	else if( strcmp(opt_name, "random") == 0 ) {
+	    if ( in_brackets && (ptr = get_bool(ptr, &i)) == NULL )
+		Handle_err;
+	    if ( i == 0 )
+		clear_random(cl->cl_option);
+	    else
+		set_random(cl->cl_option);	
+ 	    if (debug_opt)
+		fprintf(stderr, "  Opt : \"%s\" %d\n", opt_name, i);
+	}
+
+	/* handle %-line : we check if we are really in a %-line (which we do not do
+	 * for other options), because writing "&hourly" in a fcrontab results in an
+	 * error (hourly ignored) hard to find, and, in any case, annoying. */
+	else if ( cl->cl_runfreq == 1 ) {
+	    /* options to run once per interval :
+	     * ignore every fields below the limit */
+	    if (strcmp(opt_name, "mins") == 0) {
+		/* nothing to do */
+		if (debug_opt)
+		    fprintf(stderr, "  Opt : \"%s\"\n", opt_name);
+	    }
+	    else if (strcmp(opt_name, "hours") == 0) {
+		set_freq_mins(cl->cl_option);
+		if (debug_opt)
+		    fprintf(stderr, "  Opt : \"%s\"\n", opt_name);
+	    }
+	    else if (strcmp(opt_name, "days") == 0) {
+		set_freq_mins(cl->cl_option);
+		set_freq_hrs(cl->cl_option);
+		if (debug_opt)
+		    fprintf(stderr, "  Opt : \"%s\"\n", opt_name);
+	    }
+	    else if (strcmp(opt_name, "mons") == 0) {
+		set_freq_mins(cl->cl_option);
+		set_freq_hrs(cl->cl_option);
+		set_freq_days(cl->cl_option);
+		if (debug_opt)
+		    fprintf(stderr, "  Opt : \"%s\"\n", opt_name);
+	    }
+	    else if (strcmp(opt_name, "dow") == 0) {
+		set_freq_mins(cl->cl_option);
+		set_freq_hrs(cl->cl_option);
+		set_freq_days(cl->cl_option);
+		set_freq_mons(cl->cl_option);
+		if (debug_opt)
+		    fprintf(stderr, "  Opt : \"%s\"\n", opt_name);
+	    }
+
+	    /* run once an element of the selected field 
+	     * (once an hour, once a day, etc) */
+	    else if (strcmp(opt_name, "hourly") == 0) {
+		set_freq_hrs(cl->cl_option);
+		set_freq_periodically(cl->cl_option);
+		if (debug_opt)
+		    fprintf(stderr, "  Opt : \"%s\"\n", opt_name);
+	    }
+	    else if (strcmp(opt_name, "daily") == 0) {
+		set_freq_days(cl->cl_option);
+		set_freq_periodically(cl->cl_option);
+		if (debug_opt)
+		    fprintf(stderr, "  Opt : \"%s\"\n", opt_name);
+	    }
+	    else if (strcmp(opt_name, "monthly") == 0) {
+		set_freq_mons(cl->cl_option);
+		set_freq_periodically(cl->cl_option);
+		if (debug_opt)
+		    fprintf(stderr, "  Opt : \"%s\"\n", opt_name);
+	    }
+	    else if (strcmp(opt_name, "weekly") == 0) {
+		set_freq_dow(cl->cl_option);
+		set_freq_periodically(cl->cl_option);
+		if (debug_opt)
+		    fprintf(stderr, "  Opt : \"%s\"\n", opt_name);
+	    }
+
+	    /* run once an element of the selected field 
+	     * from middle to middle of that field 
+	     * (ie once from 12h to 12h the following day) */
+	    else if (strcmp(opt_name, "midhourly") == 0) {
+		set_freq_hrs(cl->cl_option);
+		set_freq_periodically(cl->cl_option);
+		set_freq_mid(cl->cl_option);
+		if (debug_opt)
+		    fprintf(stderr, "  Opt : \"%s\"\n", opt_name);
+	    }
+	    else if (strcmp(opt_name, "middaily") == 0 
+		     || strcmp(opt_name, "nightly") == 0) {
+		set_freq_days(cl->cl_option);
+		set_freq_periodically(cl->cl_option);
+		set_freq_mid(cl->cl_option);
+		if (debug_opt)
+		    fprintf(stderr, "  Opt : \"%s\"\n", opt_name);
+	    }
+	    else if (strcmp(opt_name, "midmonthly") == 0) {
+		set_freq_mons(cl->cl_option);
+		set_freq_periodically(cl->cl_option);
+		set_freq_mid(cl->cl_option);
+		if (debug_opt)
+		    fprintf(stderr, "  Opt : \"%s\"\n", opt_name);
+	    }
+	    else if (strcmp(opt_name, "midweekly") == 0) {
+		set_freq_dow(cl->cl_option);
+		set_freq_periodically(cl->cl_option);
+		set_freq_mid(cl->cl_option);
+		if (debug_opt)
+		    fprintf(stderr, "  Opt : \"%s\"\n", opt_name);
+	    }
+	}
+
 	else {
 	    fprintf(stderr, "%s:%d: Option \"%s\" unknown: "
 		    "skipping option.\n", file_name, line, opt_name);  
@@ -917,7 +981,7 @@ read_opt(char *ptr, CL *cl)
 
 
 char *
-get_time(char *ptr, time_t *time)
+get_time(char *ptr, time_t *time, int zero_allowed)
     /* convert time read in string in time_t format */
 {
     time_t sum;
@@ -929,7 +993,7 @@ get_time(char *ptr, time_t *time)
 
 	sum = 0;
 
-	while ( isdigit(*ptr) ) {
+	while ( isdigit( (int) *ptr) ) {
 	    sum *= 10;
 	    sum += *ptr - 48;
 	    ptr++;
@@ -945,6 +1009,7 @@ get_time(char *ptr, time_t *time)
 	    sum *= 24;
 	case 'h':               /* hours */
 	    sum *= 3600;
+	case 's':               /* seconds */
 	    ptr++;
 	    break;
 	case ' ':
@@ -962,7 +1027,7 @@ get_time(char *ptr, time_t *time)
     }
 
     Skip_blanks(ptr);
-    if (*time == 0) {
+    if (*time == 0 && ! zero_allowed) {
 	need_correction = 1;
 	return NULL;
     }
@@ -971,24 +1036,59 @@ get_time(char *ptr, time_t *time)
 }
 
 
+char *
+check_username(char *ptr, cf_t *cf, cl_t *cl)
+    /* check ptr to see if the first word is a username, returns new ptr */
+{
+    short int indx = 0;
+    char username[USER_NAME_LEN];
+    struct passwd *userpwent;
+
+    /* check to see if next word is a username */
+    /* we don't allow quotes, to be able to distinguish a user name from
+     * a command line (where quotes are allowed) */
+    while ( isalnum( (int) ptr[indx]) || ptr[indx] == '-' || ptr[indx] == '_' ) indx++;
+    if (indx >= USER_NAME_LEN) indx = USER_NAME_LEN - 1;
+    strncpy(username, ptr, indx);
+    username[indx] = '\0';
+
+    if ((userpwent = getpwnam(username)) != NULL) {
+	/* found the user */
+        ptr = ptr + indx;	/* move ptr to the next word */
+	Skip_blanks(ptr);
+
+	if (getuid() != ROOTUID) {
+	    fprintf(stderr, "must be privileged to run as another user : "
+		    "ignoring\n");
+	} else {
+	    Set(cl->cl_runas, username);
+            if (debug_opt)
+		fprintf(stderr, "  Opt : inline_runas %s\n", username);
+	}
+    }
+
+    return ptr;
+}
+
 
 void
-read_freq(char *ptr, CF *cf)
+read_freq(char *ptr, cf_t *cf)
     /* read a freq entry, and append a line to cf */
 {
-    CL *cl = NULL;
+    cl_t *cl = NULL;
     
-    Alloc(cl, CL);
-    memcpy(cl, &default_line, sizeof(CL));
+    Alloc(cl, cl_t);
+    memcpy(cl, &default_line, sizeof(cl_t));
     cl->cl_runas = strdup2(default_line.cl_runas);
     cl->cl_mailto = strdup2(default_line.cl_mailto);
+    cl->cl_first = -1; /* 0 is a valid value, so we have to use -1 to detect unset */
 
     /* skip the @ */
     ptr++;
 
     /* get the time before first execution or the options */
-    if ( isdigit(*ptr) ) {
-	if ( (ptr = get_time(ptr, &(cl->cl_nextexe))) == NULL ) {
+    if ( isdigit( (int) *ptr) ) {
+	if ( (ptr = get_time(ptr, &(cl->cl_first), 1)) == NULL ) {
 	    fprintf(stderr, "%s:%d: Error while reading first delay:"
 		    " skipping line.\n", file_name, line);
 	    goto exiterr;
@@ -996,7 +1096,7 @@ read_freq(char *ptr, CF *cf)
 	
 	Skip_blanks(ptr);
     }
-    else if ( isalnum(*ptr) ) {
+    else if ( isalnum( (int) *ptr) ) {
 	if ( (ptr = read_opt(ptr, cl)) == NULL )
 	    goto exiterr;
     }
@@ -1008,9 +1108,10 @@ read_freq(char *ptr, CF *cf)
     set_freq(cl->cl_option);
 
     /* then cl_timefreq */
-    if ( (ptr = get_time(ptr, &(cl->cl_timefreq))) == NULL) {
-	fprintf(stderr, "%s:%d: Error while reading frequency:"
-		" skipping line.\n", file_name, line);
+    if ( (ptr = get_time(ptr, (time_t *) &(cl->cl_timefreq), 0)) == NULL 
+	 || cl->cl_timefreq < 10 ) {
+	fprintf(stderr, "%s:%d: Error while reading frequency %s: skipping line.\n",
+		file_name, line, (cl->cl_timefreq < 10) ? "(lower than 10s) " : "");
 	goto exiterr;
     }
 	
@@ -1020,12 +1121,15 @@ read_freq(char *ptr, CF *cf)
 	goto exiterr;
     }
 
-    if ( cl->cl_nextexe == 0 )
+    if ( is_volatile(cl->cl_option) && cl->cl_first == -1 )
 	/* time before first execution is not specified */
-	cl->cl_nextexe = cl->cl_timefreq;
-
-
+	cl->cl_first = cl->cl_timefreq;
+    else
+	cl->cl_nextexe = (cl->cl_first == -1 ) ? cl->cl_timefreq : cl->cl_first;
 
+    /* check for inline runas */
+    ptr = check_username(ptr, cf, cl);
+    
     /* get cl_shell field ( remove trailing blanks ) */
     if ( (cl->cl_shell = get_string(ptr)) == NULL ) {
 	fprintf(stderr, "%s:%d: Mismatched quotes: skipping line.\n",
@@ -1048,7 +1152,7 @@ read_freq(char *ptr, CF *cf)
 
     if ( debug_opt )
 	fprintf(stderr, "  Cmd \"%s\", timefreq %ld, first %ld\n",
-		cl->cl_shell, cl->cl_timefreq, cl->cl_nextexe);
+		cl->cl_shell, cl->cl_timefreq, (long int)cl->cl_first);
     
     return;
 
@@ -1060,32 +1164,32 @@ read_freq(char *ptr, CF *cf)
 
 
 
-#define R_field(ptr, ary, max, aryconst, descrp) \
-  if((ptr = read_field(ptr, ary, max, aryconst)) == NULL) { \
+#define R_field(PTR, ARY, MAX, ARYCONST, DESCRP) \
+  if((PTR = read_field(PTR, ARY, MAX, ARYCONST)) == NULL) { \
       if (debug_opt) \
           fprintf(stderr, "\n"); \
-      fprintf(stderr, "%s:%d: Error while reading " descrp " field: " \
+      fprintf(stderr, "%s:%d: Error while reading " DESCRP " field: " \
              "skipping line.\n", file_name, line); \
       free(cl); \
       return; \
   }
 
 void
-read_arys(char *ptr, CF *cf)
+read_arys(char *ptr, cf_t *cf)
     /* read a run freq number plus a normal fcron line */
 {
-    CL *cl = NULL;
+    cl_t *cl = NULL;
     unsigned int i = 0;
 
-    Alloc(cl, CL);
-    memcpy(cl, &default_line, sizeof(CL));
+    Alloc(cl, cl_t);
+    memcpy(cl, &default_line, sizeof(cl_t));
     cl->cl_runas = strdup2(default_line.cl_runas);
     cl->cl_mailto = strdup2(default_line.cl_mailto);
 
     /* set cl_remain if not specified */
     if ( *ptr == '&' ) {
 	ptr++;
-	if ( isdigit(*ptr) ) {
+	if ( isdigit( (int) *ptr) ) {
 	    if ( (ptr = get_num(ptr, &i, USHRT_MAX, 0, NULL)) == NULL ) {
 		fprintf(stderr, "%s:%d: Error while reading runfreq:"
 			" skipping line.\n", file_name, line);
@@ -1099,7 +1203,7 @@ read_arys(char *ptr, CF *cf)
 		cl->cl_runfreq = i;
 	    }
 	}
-	else if ( isalnum(*ptr) )
+	else if ( isalnum( (int) *ptr) )
 	    if ( (ptr = read_opt(ptr, cl)) == NULL ) {
 		goto exiterr;
 	    }
@@ -1127,6 +1231,9 @@ read_arys(char *ptr, CF *cf)
 	 *  but no end line : we print it here */
 	fprintf(stderr, " remain %d\n", cl->cl_remain);
 
+    /* check for inline runas */
+    ptr = check_username(ptr, cf, cl);
+    
     /* get the shell command (remove trailing blanks) */
     if ( (cl->cl_shell = get_string(ptr)) == NULL ) {
 	fprintf(stderr, "%s:%d: Mismatched quotes: skipping line.\n",
@@ -1158,29 +1265,30 @@ read_arys(char *ptr, CF *cf)
 }
 
 void
-read_period(char *ptr, CF *cf)
+read_period(char *ptr, cf_t *cf)
     /* read a line to run periodically (i.e. once a day, once a week, etc) */
 {
-    CL *cl = NULL;
+    cl_t *cl = NULL;
     short int remain = 8;
 
-    Alloc(cl, CL);
-    memcpy(cl, &default_line, sizeof(CL));
+    Alloc(cl, cl_t);
+    memcpy(cl, &default_line, sizeof(cl_t));
     cl->cl_runas = strdup2(default_line.cl_runas);
     cl->cl_mailto = strdup2(default_line.cl_mailto);
 
     /* skip the % */
     ptr++;
 
+    /* a runfreq set to 1 means : this is a periodical line 
+     * (runfreq cannot be changed by read_opt() if already set to 1) */
+    cl->cl_remain = cl->cl_runfreq = 1;
+
     /* set cl_remain if not specified */
     if ( (ptr = read_opt(ptr, cl)) == NULL ) {
 	goto exiterr;
     }
     Skip_blanks(ptr);
 
-    /* a runfreq set to 1 means : this is a periodical line */
-    cl->cl_remain = cl->cl_runfreq = 1;
-
     /* we set this here, because it may be unset by read_opt (reset option) */
     set_td(cl->cl_option);
 
@@ -1213,6 +1321,9 @@ read_period(char *ptr, CF *cf)
 	 *  but no end line : we print it here */
 	fprintf(stderr, " remain %d\n", cl->cl_remain);
 
+    /* check for inline runas */
+    ptr = check_username(ptr, cf, cl);
+    
     /* get the shell command (remove trailing blanks) */
     if ( (cl->cl_shell = get_string(ptr)) == NULL ) {
 	fprintf(stderr, "%s:%d: Mismatched quotes: skipping line.\n",
@@ -1224,7 +1335,7 @@ read_period(char *ptr, CF *cf)
 		file_name, line);
 	goto exiterr;
     } 
-    else if ( cl->cl_shell[0] == '*' || isdigit(cl->cl_shell[0]) )
+    else if ( cl->cl_shell[0] == '*' || isdigit( (int) cl->cl_shell[0]) )
 	fprintf(stderr, "%s:%d: Warning : shell command beginning by '%c'.\n",
 		file_name, line, cl->cl_shell[0]);
 
@@ -1297,7 +1408,7 @@ get_num(char *ptr, int *num, int max, short int decimal, const char **names)
     int i = 0;
     *num = 0;
 
-    if ( isalpha(*ptr) ) {
+    if ( isalpha( (int) *ptr) ) {
 
 	if ( names == NULL ) {
 	    need_correction = 1;
@@ -1322,7 +1433,7 @@ get_num(char *ptr, int *num, int max, short int decimal, const char **names)
 
     } else {
 
-	while ( isdigit(*ptr) || *ptr == '.') { 
+	while ( isdigit( (int) *ptr) || *ptr == '.') { 
 
 	    if ( *ptr == '.' && ptr++ && i++ > 0 )
 		return NULL;
@@ -1331,7 +1442,7 @@ get_num(char *ptr, int *num, int max, short int decimal, const char **names)
 		 * skip the other decimals and return */
 		if ( *ptr >= '5' )
 		    *num += 1;
-		while ( isdigit(*(++ptr)) ) ;
+		while ( isdigit( (int) *(++ptr)) ) ;
 		ptr--;
 	    } else {
 		*num *= 10; 
@@ -1487,10 +1598,10 @@ delete_file(const char *user_name)
     /* free a file if user_name is not null 
      *   otherwise free all files */
 {
-    CF *file = NULL;
-    CF *prev_file = NULL;
-    CL *line = NULL;
-    CL *cur_line = NULL;
+    cf_t *file = NULL;
+    cf_t *prev_file = NULL;
+    cl_t *line = NULL;
+    cl_t *cur_line = NULL;
     env_t *env = NULL;
     env_t *cur_env = NULL;
 
@@ -1539,58 +1650,27 @@ delete_file(const char *user_name)
 
 }
 
-
-void
+int
 save_file(char *path)
     /* Store the informations relatives to the executions
      * of tasks at a defined frequency of system's running time */
 {
-    CF *file = NULL;
-    CL *line = NULL;
-    FILE *f = NULL;
-    env_t *env = NULL;
+    cf_t *file = NULL;
 
     if (debug_opt)
 	fprintf(stderr, "Saving ...\n");
     
     for (file = file_base; file; file = file->cf_next) {
 
-	/* open file */
-	if ( (f = fopen(path, "w")) == NULL )
-	    die_e("File has not be saved");
-
-	/* save file : */
-
-	/* put program's version : it permit to daemon not to load
-	 * a file which he won't understand the syntax, for exemple
-	 * a file using a depreciated format generated by an old fcrontab,
-	 * if the syntax has changed */
-	fprintf(f, "fcrontab-" FILEVERSION "\n");
-
-	/* put the user's name : it is used to check his uid has not changed */
- 	fprintf(f, "%s%c", user, '\0');
-
-	/* put the time & date of saving : this is use for calcutating 
-	 * the system down time. As it is a new file, we set it to 0 */
-	fprintf(f, "%d", 0);
-
-	/*   env variables, */
-	for (env = file->cf_env_base; env; env = env->e_next)
-	    fprintf(f, "%s%c", env->e_val, '\0');
-	
-	fprintf(f, "%c", '\0');
-
-	/*   then, lines. */
-	for (line = file->cf_line_base; line; line = line->cl_next) {
-	    if ( fwrite(line, sizeof(CL), 1, f) != 1 )
-		perror("save");
-	    fprintf(f, "%s%c", line->cl_shell, '\0');
-	    fprintf(f, "%s%c", line->cl_runas, '\0');
-	    fprintf(f, "%s%c", line->cl_mailto, '\0');
-	}
-
-	fclose(f);
+    /* save_file() is run under user's rights.
+     * If fcrontab is run by root for a normal user, we must change the file's
+     * ownership to this user, in order to make fcron check the runas fields.
+     * (a malicious user could put a runas(root) and wait for the fcrontab to be
+     * installed by root) */
+	if ( save_file_safe(file, path, "fcrontab", asuid, fcrontab_gid, 0) == ERR )
+	    return ERR;
 
     }
-}
 
+    return OK;
+}
